from .__version__ import __version__
from .base import FAAPI
from .comment import Comment
from .journal import Journal
from .journal import JournalPartial
from .submission import Submission
from .submission import SubmissionPartial
from .user import User
from .user import UserPartial

__all__ = [
    "__version__",
    "FAAPI",
    "Comment",
    "Journal",
    "JournalPartial",
    "Submission",
    "SubmissionPartial",
    "User",
    "UserPartial",
    "exceptions",
    "connection",
    "parse"
]



from datetime import datetime
from typing import Optional
from typing import Union

from bs4.element import Tag

import faapi
from .exceptions import _raise_exception
from .parse import html_to_bbcode
from .parse import parse_comment_tag


class Comment:
    """
    Contains comment information and references to replies and parent objects.
    """

    def __init__(self, tag: Tag = None, parent: Union[faapi.submission.Submission, faapi.journal.Journal] = None):
        """
        :param tag: The comment tag from which to parse information
        :param parent: The parent object of the comment
        """
        assert tag is None or isinstance(tag, Tag), _raise_exception(TypeError(f"tag must be {None} or {Tag.__name__}"))

        self.comment_tag: Optional[Tag] = tag

        self.id: int = 0
        self.author: faapi.user.UserPartial = faapi.user.UserPartial()
        self.date: datetime = datetime.fromtimestamp(0)
        self.text: str = ""
        self.replies: list[Comment] = []
        self.reply_to: Optional[Union[Comment, int]] = None
        self.edited: bool = False
        self.hidden: bool = False
        self.parent: Optional[Union[faapi.submission.Submission, faapi.journal.Journal]] = parent

        self.parse()

    def __hash__(self) -> int:
        return hash((self.id, type(self.parent), self.parent))

    def __eq__(self, other) -> bool:
        if isinstance(other, Comment):
            return other.id == self.id and self.parent == other.parent
        elif isinstance(other, int):
            return other == self.id
        return False

    def __gt__(self, other) -> bool:
        if isinstance(other, Comment):
            return self.id > other.id
        elif isinstance(other, int):
            return self.id > other
        return False

    def __ge__(self, other) -> bool:
        if isinstance(other, Comment):
            return self.id >= other.id
        elif isinstance(other, int):
            return self.id >= other
        return False

    def __lt__(self, other) -> bool:
        if isinstance(other, Comment):
            return self.id < other.id
        elif isinstance(other, int):
            return self.id < other
        return False

    def __le__(self, other) -> bool:
        if isinstance(other, Comment):
            return self.id <= other.id
        elif isinstance(other, int):
            return self.id <= other
        return False

    def __iter__(self):
        yield "id", self.id
        yield "author", dict(self.author)
        yield "date", self.date
        yield "text", self.text
        yield "replies", [dict(r) for r in self.replies]
        yield "reply_to", dict(_remove_recursion(self.reply_to)) if isinstance(self.reply_to, Comment) \
            else self.reply_to
        yield "edited", self.edited
        yield "hidden", self.hidden
        yield "parent", None if self.parent is None else dict(self.parent)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"{self.id} {self.author}".rstrip()

    @property
    def text_bbcode(self) -> str:
        """
        The comment text formatted to BBCode

        :return: BBCode text
        """
        return html_to_bbcode(self.text)

    @property
    def url(self):
        """
        Compose the full URL to the comment.

        :return: The URL to the comment.
        """
        return "" if self.parent is None else f"{self.parent.url}#cid:{self.id}"

    def parse(self, comment_tag: Tag = None):
        """
        Parse a comment tag, overrides any information already present in the object.

        :param comment_tag: The comment tag from which to parse information
        """
        assert comment_tag is None or isinstance(comment_tag, Tag), \
            _raise_exception(TypeError(f"tag must be {None} or {Tag.__name__}"))

        self.comment_tag = comment_tag or self.comment_tag
        if self.comment_tag is None:
            return

        parsed: dict = parse_comment_tag(self.comment_tag)

        self.id = parsed["id"]
        self.date = datetime.fromtimestamp(parsed["timestamp"])
        self.author = faapi.user.UserPartial()
        self.author.name = parsed["user_name"]
        self.author.title = parsed["user_title"]
        self.author.user_icon_url = parsed["user_icon_url"]
        self.text = parsed["text"]
        self.replies = []
        self.reply_to = parsed["parent"]
        self.edited = parsed["edited"]
        self.hidden = parsed["hidden"]


def sort_comments(comments: list[Comment]) -> list[Comment]:
    """
    Sort a list of comments into a tree structure. Replies are overwritten.

    :param comments: A list of Comment objects (flat or tree-structured)
    :return: A tree-structured list of comments with replies
    """
    for comment in (comments := sorted(flatten_comments(comments))):
        comment.replies = [_set_reply_to(c, comment) for c in comments if c.reply_to == comment]
    return [c for c in comments if c.reply_to is None]


def flatten_comments(comments: list[Comment]) -> list[Comment]:
    """
    Flattens a list of comments. Replies are not modified.

    :param comments: A list of Comment objects (flat or tree-structured)
    :return: A flat date-sorted (ascending) list of comments
    """
    return sorted({c for c in [r for c in comments for r in [c, *flatten_comments(c.replies)]]})


def _set_reply_to(comment: Comment, reply_to: Union[Comment, int]) -> Comment:
    comment.reply_to = reply_to
    return comment


def _remove_recursion(comment: Comment) -> Comment:
    comment_new: Comment = Comment()

    comment_new.comment_tag = comment.comment_tag
    comment_new.id = comment.id
    comment_new.author = comment.author
    comment_new.date = comment.date
    comment_new.text = comment.text
    comment_new.replies = [_remove_recursion(c) for c in comment.replies]
    comment_new.reply_to = comment.reply_to.id if isinstance(comment.reply_to, Comment) else comment.reply_to
    comment_new.edited = comment.edited
    comment_new.hidden = comment.hidden
    comment_new.parent = None

    return comment_new



from http.cookiejar import CookieJar
from time import sleep
from time import time
from typing import Any
from typing import Optional
from typing import Union
from urllib.robotparser import RobotFileParser

from .connection import CloudflareScraper
from .connection import CookieDict
from .connection import Response
from .connection import get
from .connection import get_robots
from .connection import join_url
from .connection import make_session
from .connection import stream_binary
from .exceptions import DisallowedPath
from .exceptions import Unauthorized
from .journal import Journal
from .journal import JournalPartial
from .parse import BeautifulSoup
from .parse import check_page_raise
from .parse import parse_loggedin_user
from .parse import parse_page
from .parse import parse_submission_figures
from .parse import parse_user_favorites
from .parse import parse_user_journals
from .parse import parse_user_submissions
from .parse import parse_watchlist
from .parse import username_url
from .submission import Submission
from .submission import SubmissionPartial
from .user import User
from .user import UserPartial


# noinspection GrazieInspection
class FAAPI:
    """
    This class provides the methods to access and parse Fur Affinity pages and retrieve objects.
    """

    def __init__(self, cookies: Union[list[CookieDict], CookieJar]):
        """
        :param cookies: The cookies for the session.
        """

        self.session: CloudflareScraper = make_session(cookies)  # Session used for get requests
        self.robots: RobotFileParser = get_robots(self.session)  # robots.txt handler
        self.last_get: float = time() - self.crawl_delay  # Time of last get (UNIX time)
        self.raise_for_unauthorized: bool = True  # Control login checks
        self.timeout: Optional[int] = None  # Timeout for requests

    @property
    def user_agent(self) -> str:
        """
        The user agent of the session
        """
        return self.session.headers["User-Agent"]

    @property
    def crawl_delay(self) -> float:
        """
        Crawl delay from robots.txt
        """
        return float(self.robots.crawl_delay(self.user_agent) or 1)

    def load_cookies(self, cookies: Union[list[CookieDict], CookieJar]):
        """
        Load new cookies and create a new session.

        :param cookies: The cookies for the session.
        """
        self.session = make_session(cookies)

    def handle_delay(self):
        """
        Handles the crawl delay as set in the robots.txt
        """
        if (d := time() - self.last_get) < self.crawl_delay:
            sleep(self.crawl_delay - d)
        self.last_get = time()

    def check_path(self, path: str, *, raise_for_disallowed: bool = False) -> bool:
        """
        Checks whether a given path is allowed by the robots.txt.

        :param path: The path to check.
        :param raise_for_disallowed: Whether to raise an exception for a non-allowed path.
        :return: True if the path is allowed in the robots.txt, False otherwise.
        """
        if not (allowed := self.robots.can_fetch(self.user_agent, "/" + path.lstrip("/"))) and raise_for_disallowed:
            raise DisallowedPath(f"Path {path!r} is not allowed by robots.txt")
        return allowed

    @property
    def connection_status(self) -> bool:
        """
        Check the status of the connection to Fur Affinity.

        :return: True if it can connect, False otherwise.
        """
        try:
            return self.get("/").ok
        except ConnectionError:
            return False

    @property
    def login_status(self) -> bool:
        """
        Check the login status of the given cookies.

        :return: True if the cookies belong to a login session, False otherwise.
        """
        return parse_loggedin_user(self.get_parsed("login", skip_auth_check=True)) is not None

    def get(self, path: str, **params: Union[str, bytes, int, float]) -> Response:
        """
        Fetch a path with a GET request.
        The path is checked against the robots.txt before the request is made.
        The crawl-delay setting is enforced wth a wait time.

        :param path: The path to fetch.
        :param params: Query parameters for the request.
        :return: A Response object from the request.
        """
        self.check_path(path, raise_for_disallowed=True)
        self.handle_delay()
        return get(self.session, path, timeout=self.timeout, params=params)

    def get_parsed(self, path: str, *, skip_page_check: bool = False, skip_auth_check: bool = False,
                   **params: Union[str, bytes, int, float]) -> BeautifulSoup:
        """
        Fetch a path with a GET request and parse it using BeautifulSoup.

        :param path: The path to fetch.
        :param skip_page_check: Whether to skip checking the parsed page for errors.
        :param skip_auth_check: Whether to skip checking the parsed page for login status.
        :param params: Query parameters for the request.
        :return: A BeautifulSoup object containing the parsed content of the request response.
        """
        response: Response = self.get(path, **params)
        response.raise_for_status()
        page: BeautifulSoup = parse_page(response.text)
        if not skip_page_check:
            check_page_raise(page)
        if not skip_auth_check and self.raise_for_unauthorized and not parse_loggedin_user(page):
            raise Unauthorized("Not logged in")
        return page

    def me(self) -> Optional[User]:
        """
        Fetch the information of the logged-in user.

        :return: A User object for the logged-in user, or None if the cookies are not from a login session.
        """
        return self.user(user) if (user := parse_loggedin_user(self.get_parsed("login"))) else None

    def frontpage(self) -> list[SubmissionPartial]:
        """
        Fetch latest submissions from Fur Affinity's front page

        :return: A list of SubmissionPartial objects
        """
        page_parsed: BeautifulSoup = self.get_parsed("/")
        submissions: list[SubmissionPartial] = [SubmissionPartial(f) for f in parse_submission_figures(page_parsed)]
        return sorted({s for s in submissions}, reverse=True)

    def submission(self, submission_id: int, get_file: bool = False, *, chunk_size: int = None
                   ) -> tuple[Submission, Optional[bytes]]:
        """
        Fetch a submission and, optionally, its file.

        :param submission_id: The ID of the submission.
        :param get_file: Whether to download the submission file.
        :param chunk_size: The chunk_size to be used for the download (does not override get_file).
        :return: A Submission object and a bytes object (if the submission file is downloaded).
        """
        sub: Submission = Submission(self.get_parsed(join_url("view", int(submission_id))))
        sub_file: Optional[bytes] = self.submission_file(sub, chunk_size=chunk_size) if get_file and sub.id else None
        return sub, sub_file

    def submission_file(self, submission: Submission, *, chunk_size: int = None) -> bytes:
        """
        Fetch a submission file from a Submission object.

        :param submission: A Submission object.
        :param chunk_size: The chunk_size to be used for the download.
        :return: The submission file as a bytes object.
        """
        self.handle_delay()
        return stream_binary(self.session, submission.file_url, chunk_size=chunk_size, timeout=self.timeout)

    def journal(self, journal_id: int) -> Journal:
        """
        Fetch a journal.

        :param journal_id: The ID of the journal.
        :return: A Journal object.
        """
        return Journal(self.get_parsed(join_url("journal", int(journal_id))))

    def user(self, user: str) -> User:
        """
        Fetch a user.

        :param user: The name of the user (_ characters are allowed).
        :return: A User object.
        """
        return User(self.get_parsed(join_url("user", username_url(user))))

    # noinspection DuplicatedCode
    def gallery(self, user: str, page: int = 1) -> tuple[list[SubmissionPartial], Optional[int]]:
        """
        Fetch a user's gallery page.

        :param user: The name of the user (_ characters are allowed).
        :param page: The page to fetch.
        :return: A list of SubmissionPartial objects and the next page (None if it is the last).
        """
        page_parsed: BeautifulSoup = self.get_parsed(join_url("gallery", username_url(user), int(page)))
        info_parsed: dict[str, Any] = parse_user_submissions(page_parsed)
        author: UserPartial = UserPartial()
        author.name, author.status, author.title, author.join_date, author.user_icon_url = [
            info_parsed["user_name"], info_parsed["user_status"],
            info_parsed["user_title"], info_parsed["user_join_date"],
            info_parsed["user_icon_url"]
        ]
        for s in (submissions := list(map(SubmissionPartial, info_parsed["figures"]))):
            s.author = author
        return submissions, (page + 1) if not info_parsed["last_page"] else None

    # noinspection DuplicatedCode
    def scraps(self, user: str, page: int = 1) -> tuple[list[SubmissionPartial], Optional[int]]:
        """
        Fetch a user's scraps page.

        :param user: The name of the user (_ characters are allowed).
        :param page: The page to fetch.
        :return: A list of SubmissionPartial objects and the next page (None if it is the last).
        """
        page_parsed: BeautifulSoup = self.get_parsed(join_url("scraps", username_url(user), int(page)))
        info_parsed: dict[str, Any] = parse_user_submissions(page_parsed)
        author: UserPartial = UserPartial()
        author.name, author.status, author.title, author.join_date, author.user_icon_url = [
            info_parsed["user_name"], info_parsed["user_status"],
            info_parsed["user_title"], info_parsed["user_join_date"],
            info_parsed["user_icon_url"]
        ]
        for s in (submissions := list(map(SubmissionPartial, info_parsed["figures"]))):
            s.author = author
        return submissions, (page + 1) if not info_parsed["last_page"] else None

    def favorites(self, user: str, page: str = "") -> tuple[list[SubmissionPartial], Optional[str]]:
        """
        Fetch a user's favorites page.

        :param user: The name of the user (_ characters are allowed).
        :param page: The page to fetch.
        :return: A list of SubmissionPartial objects and the next page (None if it is the last).
        """
        page_parsed: BeautifulSoup = self.get_parsed(join_url("favorites", username_url(user), page.strip()))
        info_parsed: dict[str, Any] = parse_user_favorites(page_parsed)
        submissions: list[SubmissionPartial] = list(map(SubmissionPartial, info_parsed["figures"]))
        return submissions, info_parsed["next_page"] or None

    def journals(self, user: str, page: int = 1) -> tuple[list[JournalPartial], Optional[int]]:
        """
        Fetch a user's journals page.

        :param user: The name of the user (_ characters are allowed).
        :param page: The page to fetch.
        :return: A list of Journal objects and the next page (None if it is the last).
        """
        page_parsed: BeautifulSoup = self.get_parsed(join_url("journals", username_url(user), int(page)))
        info_parsed: dict[str, Any] = parse_user_journals(page_parsed)
        author: UserPartial = UserPartial()
        author.name, author.status, author.title, author.join_date, author.user_icon_url = [
            info_parsed["user_name"], info_parsed["user_status"],
            info_parsed["user_title"], info_parsed["user_join_date"],
            info_parsed["user_icon_url"]
        ]
        for j in (journals := list(map(JournalPartial, info_parsed["sections"]))):
            j.author = author
        return journals, (page + 1) if not info_parsed["last_page"] else None

    def watchlist_to(self, user: str, page: int = 1) -> tuple[list[UserPartial], Optional[int]]:
        """
        Fetch a page from the list of users watching the user.

        :param user: The name of the user (_ characters are allowed).
        :param page: The page to fetch.
        :return: A list of UserPartial objects and the next page (None if it is the last).
        """
        users: list[UserPartial] = []
        us, np = parse_watchlist(
            self.get_parsed(join_url("watchlist", "to", username_url(user), page), skip_auth_check=True))
        for s, u in us:
            _user: UserPartial = UserPartial()
            _user.name = u
            _user.status = s
            users.append(_user)
        return users, np if np and np != page else None

    def watchlist_by(self, user: str, page: int = 1) -> tuple[list[UserPartial], Optional[int]]:
        """
        Fetch a page from the list of users watched by the user.
        :param user: The name of the user (_ characters are allowed).
        :param page: The page to fetch.
        :return: A list of UserPartial objects and the next page (None if it is the last).
        """
        users: list[UserPartial] = []
        us, np = parse_watchlist(
            self.get_parsed(join_url("watchlist", "by", username_url(user), page), skip_auth_check=True))
        for s, u in us:
            _user: UserPartial = UserPartial()
            _user.name = u
            _user.status = s
            users.append(_user)
        return users, np if np and np != page else None



from http.client import IncompleteRead
from http.cookiejar import Cookie
from http.cookiejar import CookieJar
from platform import python_version
from platform import uname
from re import compile as re_compile
from typing import Optional
from typing import TypedDict
from typing import Union
from urllib.robotparser import RobotFileParser

from cfscrape import CloudflareScraper  # type: ignore
from cfscrape import create_scraper  # type: ignore
from requests import Response
from requests import Session

from .__version__ import __version__
from .exceptions import Unauthorized
from .exceptions import _raise_exception

root: str = "https://www.furaffinity.net"


class CookieDict(TypedDict):
    name: str
    value: str


def join_url(*url_comps: Union[str, int]) -> str:
    return "/".join(map(lambda e: str(e).strip(" /"), url_comps))


def make_session(cookies: Union[list[CookieDict], CookieJar]) -> CloudflareScraper:
    assert len(cookies), _raise_exception(Unauthorized("No cookies for session"))
    session: CloudflareScraper = create_scraper()
    session.headers["User-Agent"] = f"faapi/{__version__} Python/{python_version()} {(u := uname()).system}/{u.release}"

    for cookie in cookies:
        if isinstance(cookie, Cookie):
            session.cookies.set(cookie.name, cookie.value)
        else:
            session.cookies.set(cookie["name"], cookie["value"])

    return session


def get_robots(session: Session) -> RobotFileParser:
    robots: RobotFileParser = RobotFileParser(url := join_url(root, "robots.txt"))
    robots.parse(filter(re_compile(r"^[^#\s].+").match, map(str.strip, session.get(url).text.splitlines())))
    return robots


def get(session: CloudflareScraper, path: str, *, timeout: int = None,
        params: dict[str, Union[str, bytes, int, float]] = None) -> Response:
    return session.get(join_url(root, path), params=params, timeout=timeout)


def stream_binary(session: CloudflareScraper, url: str, *, chunk_size: Optional[int] = None,
                  timeout: Optional[int] = None) -> bytes:
    stream: Response = session.get(url, stream=True, timeout=timeout)
    stream.raise_for_status()

    file_binary: bytes = bytes().join(stream.iter_content(chunk_size))

    if (length := int(stream.headers.get("Content-Length", 0))) > 0 and length != len(file_binary):
        raise IncompleteRead(file_binary, length - len(file_binary))

    return file_binary



from datetime import datetime
from re import MULTILINE
from re import Match
from re import Pattern
from re import compile as re_compile
from re import match
from re import search
from re import sub
from typing import Any
from typing import Optional
from typing import Union

from bbcode import Parser as BBCodeParser  # type:ignore
from bs4 import BeautifulSoup
from bs4.element import NavigableString
from bs4.element import Tag
from dateutil.parser import parse as parse_date
from htmlmin import minify  # type:ignore

from .connection import root
from .exceptions import DisabledAccount
from .exceptions import NoTitle
from .exceptions import NonePage
from .exceptions import NotFound
from .exceptions import NoticeMessage
from .exceptions import ParsingError
from .exceptions import ServerError
from .exceptions import _raise_exception

relative_url: Pattern = re_compile(r"^(?:https?://(?:www\.)?furaffinity\.net)?(.*)")
mentions_regexp: Pattern = re_compile(r"^(?:(?:https?://)?(?:www\.)?furaffinity\.net)?/user/([^/#]+).*$")
watchlist_next_regexp: Pattern = re_compile(r"/watchlist/(?:by|to)/[^/]+/(\d+)")
not_found_messages: tuple[str, ...] = ("not in our database", "cannot be found", "could not be found", "user not found")
deactivated_messages: tuple[str, ...] = ("deactivated", "pending deletion")
smilie_icons: tuple[str, ...] = (
    "crying", "derp", "dunno", "embarrassed", "evil", "gift", "huh", "lmao", "love", "nerd", "note", "oooh", "pleased",
    "rollingeyes", "sad", "sarcastic", "serious", "sleepy", "smile", "teeth", "tongue", "veryhappy", "wink", "yelling",
    "zipped", "angel", "badhairday", "cd", "coffee", "cool", "whatever"
)


def get_attr(tag: Tag, attr: str) -> str:
    return value[0] if isinstance(value := tag.attrs[attr], list) else value


def parse_page(text: str) -> BeautifulSoup:
    return BeautifulSoup(text, "lxml")


def check_page_raise(page: BeautifulSoup) -> None:
    if page is None:
        raise NonePage
    elif not (title := page.title.text.lower() if page.title else ""):
        raise NoTitle
    elif title.startswith("account disabled"):
        raise DisabledAccount
    elif title == "system error":
        error_text: str = error.text if (error := page.select_one("div.section-body")) else ""
        if any(m in error_text.lower() for m in not_found_messages):
            raise NotFound
        else:
            raise ServerError(*filter(bool, map(str.strip, error_text.splitlines())))
    elif notice := page.select_one("section.notice-message"):
        notice_text: str = notice.text
        if any(m in notice_text.lower() for m in deactivated_messages):
            raise DisabledAccount
        elif any(m in notice_text.lower() for m in not_found_messages):
            raise NotFound
        else:
            raise NoticeMessage(*filter(bool, map(str.strip, notice_text.splitlines())))


def username_url(username: str) -> str:
    return sub(r"[^a-z\d.~-]", "", username.lower())


def inner_html(tag: Tag) -> str:
    return tag.decode_contents()


def clean_html(html: str) -> str:
    return sub(r" *(<br/?>) *", r"\1", minify(html, remove_comments=True, reduce_boolean_attributes=True)).strip()


def html_to_bbcode(html: str) -> str:
    body: Optional[Tag] = parse_page(f"<html><body>{html}</body></html>").select_one("html > body")
    if not body:
        return ""

    for linkusername in body.select("a.linkusername"):
        linkusername.replaceWith(f"@{linkusername.text.strip()}")

    for iconusername in body.select("a.iconusername,a.usernameicon"):
        username: str = iconusername.text.strip() or iconusername.attrs.get('href', '').strip('/').split('/')[-1]
        if icon := iconusername.select_one("img"):
            username = icon.attrs.get('alt', '').strip() or username
        iconusername.replaceWith(f":icon{username}:" if iconusername.text.strip() else f":{username}icon:")

    for img in body.select("img"):
        img.replaceWith(f"[img={img.attrs.get('src', '')}/]")

    for hr in body.select("hr"):
        hr.replaceWith("-----")

    for smilie in body.select("i.smilie"):
        smilie_class: list[str] = list(smilie.attrs.get("class", []))
        smilie_name: str = next(filter(lambda c: c not in ["smilie", ""], smilie_class), "")
        smilie.replaceWith(f":{smilie_name or 'smilie'}:")

    for span in body.select("span.bbcode[style*=color]"):
        if m := match(r".*color: ?([^ ;]+).*", span.attrs["style"]):
            span.replaceWith(f"[color={m[1]}]", *span.children, "[/color]")
        else:
            span.replaceWith(*span.children)

    for nav_link in body.select("span.parsed_nav_links"):
        a_tags = nav_link.select("a")
        a_prev_tag: Optional[Tag] = next((a for a in a_tags if "prev" in a.text.lower()), None)
        a_frst_tag: Optional[Tag] = next((a for a in a_tags if "first" in a.text.lower()), None)
        a_next_tag: Optional[Tag] = next((a for a in a_tags if "next" in a.text.lower()), None)
        a_prev = a_prev_tag.attrs.get("href", "").strip("/").split("/")[-1] if a_prev_tag else ""
        a_frst = a_frst_tag.attrs.get("href", "").strip("/").split("/")[-1] if a_frst_tag else ""
        a_next = a_next_tag.attrs.get("href", "").strip("/").split("/")[-1] if a_next_tag else ""
        nav_link.replaceWith(f"[{a_prev or '-'},{a_frst or '-'},{a_next or '-'}]")

    for a in body.select("a.auto_link_shortened:not(.named_url), a.auto_link:not(.named_url)"):
        a.replaceWith(a.attrs.get('href', ''))

    for a in body.select("a"):
        href_match: Optional[Match] = relative_url.match(a.attrs.get('href', ''))
        a.replaceWith(f"[url={href_match[1] if href_match else a.attrs.get('href', '')}]",
                      *a.children,
                      "[/url]")

    for yt in body.select("iframe[src*='youtube.com/embed']"):
        yt.replaceWith(f"[yt]https://youtube.com/embed/{yt.attrs.get('src', '').strip('/').split('/')}[/yt]")

    for quote_name_tag in body.select("span.bbcode.bbcode_quote > span.bbcode_quote_name"):
        quote_author: str = quote_name_tag.text.strip().removesuffix('wrote:').strip()
        quote_tag = quote_name_tag.parent
        if not quote_tag:
            quote_name_tag.replaceWith(quote_author)
            continue
        quote_name_tag.decompose()
        quote_tag.replaceWith(f"[quote{('=' + quote_author) if quote_author else ''}]",
                              *quote_tag.children,
                              "[/quote]")

    for quote_tag in body.select("span.bbcode.bbcode_quote"):
        quote_tag.replaceWith("[quote]", *quote_tag.children, "[/quote]")

    for [selector, bbcode_tag] in (
            ("i", "i"),
            ("b", "b"),
            ("strong", "b"),
            ("u", "u"),
            ("s", "s"),
            ("code.bbcode_left", "left"),
            ("code.bbcode_center", "center"),
            ("code.bbcode_right", "right"),
            ("span.bbcode_spoiler", "spoiler"),
            ("sub", "sub"),
            ("sup", "sup"),
            ("h1", "h1"),
            ("h2", "h2"),
            ("h3", "h3"),
            ("h4", "h4"),
            ("h5", "h5"),
            ("h6", "h6"),
    ):
        for tag in body.select(selector):
            tag.replaceWith(f"[{bbcode_tag}]", *tag.children, f"[/{bbcode_tag}]")

    for br in body.select("br"):
        br.replaceWith("\n")

    for p in body.select("p"):
        p.replaceWith(*p.children)

    for tag in body.select("*"):
        if not (div_class := tag.attrs.get("class", None)):
            tag.replaceWith(f"[tag={tag.name}]", *tag.children, "[/tag.{tag.name}]")
        else:
            tag.replaceWith(f"[tag={tag.name}.{' '.join(div_class) if isinstance(div_class, list) else div_class}]",
                            *tag.children,
                            "[/tag]")

    bbcode: str = body.decode_contents()

    bbcode = sub(" *$", "", bbcode, flags=MULTILINE)
    bbcode = sub("^ *", "", bbcode, flags=MULTILINE)

    for char, substitution in (
            ("©", "(c)"),
            ("™", "(tm)"),
            ("®", "(r)"),
            ("&copy;", "(c)"),
            ("&reg;", "(tm)"),
            ("&trade;", "(r)"),
            ("&lt;", "<"),
            ("&gt;", ">"),
            ("&amp;", "&"),
    ):
        bbcode = bbcode.replace(char, substitution)

    return bbcode.strip(" ")


def bbcode_to_html(bbcode: str) -> str:
    def render_url(_tag_name, value: str, options: dict[str, str], _parent, _context) -> str:
        return f'<a class="auto_link named_url" href="{options.get("url", "#")}">{value}</a>'

    def render_color(_tag_name, value, options, _parent, _context) -> str:
        return f'<span class=bbcode style="color:{options.get("color", "inherit")};">{value}</span>'

    def render_quote(_tag_name, value: str, options: dict[str, str], _parent, _context) -> str:
        author: str = options.get("quote", "")
        author = f"<span class=bbcode_quote_name>{author} wrote:</span>" if author else ""
        return f'<span class="bbcode bbcode_quote">{author}{value}</span>'

    def render_tags(tag_name: str, value: str, options: dict[str, str], _parent, _context) -> str:
        if not options and tag_name.islower():
            return f"<{tag_name}>{value}</{tag_name}>"
        return f"[{tag_name} {' '.join(f'{k}={v}' if v else k for k, v in options.items())}]{value}"

    def render_tag(_tag_name, value: str, options: dict[str, str], _parent, _context) -> str:
        name, *classes = options["tag"].split(".")
        return f'<{name} class="{" ".join(classes)}">{value}</{name}>'

    def parse_extra(page: BeautifulSoup) -> BeautifulSoup:
        child: NavigableString
        child_new: Tag
        has_match: bool = True
        while has_match:
            has_match = False
            for child in [c for e in page.select("*:not(a)") for c in e.children if isinstance(c, NavigableString)]:
                if m_ := match(rf"(.*):({'|'.join(smilie_icons)}):(.*)", child):
                    has_match = True
                    child_new = Tag(name="i", attrs={"class": f"smilie {m_[2]}"})
                    child.replaceWith(m_[1], child_new, m_[3])
                elif m_ := match(r"(.*)(?:@([a-zA-Z0-9.~_-]+)|:link([a-zA-Z0-9.~_-]+):)(.*)", child):
                    has_match = True
                    child_new = Tag(name="a", attrs={"class": "linkusername", "href": f"/user/{m_[2] or m_[3]}"})
                    child_new.insert(0, m_[2] or m_[3])
                    child.replaceWith(m_[1], child_new, m_[4])
                elif m_ := match(r"(.*):(?:icon([a-zA-Z0-9.~_-]+)|([a-zA-Z0-9.~_-]+)icon):(.*)", child):
                    has_match = True
                    user: str = m_[2] or m_[3] or ""
                    child_new = Tag(name="a", attrs={"class": "iconusername", "href": f"/user/{user}"})
                    child_new_img: Tag = Tag(name="img",
                                             attrs={"alt": user, "title": user,
                                                    "src": f"//a.furaffinity.net/{datetime.now():%Y%m%d}/{user}.gif"})
                    child_new.insert(0, child_new_img)
                    if m_[2]:
                        child_new.insert(1, f"\xA0{m_[2]}")
                    child.replaceWith(m_[1], child_new, m_[4])
                elif m_ := match(r"(.*)\[ *(?:(\d+)|-)?, *(?:(\d+)|-)? *, *(?:(\d+)|-)? *](.*)", child):
                    has_match = True
                    child_new = Tag(name="span", attrs={"class": "parsed_nav_links"})
                    child_new_1: Union[Tag, str] = "<<<\xA0PREV"
                    child_new_2: Union[Tag, str] = "FIRST"
                    child_new_3: Union[Tag, str] = "NEXT\xA0>>>"
                    if m_[2]:
                        child_new_1 = Tag(name="a", attrs={"href": f"/view/{m_[2]}"})
                        child_new_1.insert(0, "<<<\xA0PREV")
                    if m_[3]:
                        child_new_2 = Tag(name="a", attrs={"href": f"/view/{m_[3]}"})
                        child_new_2.insert(0, "<<<\xA0FIRST")
                    if m_[4]:
                        child_new_3 = Tag(name="a", attrs={"href": f"/view/{m_[4]}"})
                        child_new_3.insert(0, "NEXT\xA0>>>")
                    child_new.insert(0, child_new_1)
                    child_new.insert(1, "\xA0|\xA0")
                    child_new.insert(2, child_new_2)
                    child_new.insert(3, "\xA0|\xA0")
                    child_new.insert(4, child_new_3)
                    child.replaceWith(m_[1], child_new, m_[5])

        for p in page.select("p"):
            p.replaceWith(*p.children)

        return page

    parser: BBCodeParser = BBCodeParser(install_defaults=False, replace_links=False, replace_cosmetic=True)
    parser.REPLACE_ESCAPE = (
        ("&", "&amp;"),
        ("<", "&lt;"),
        (">", "&gt;"),
    )
    parser.REPLACE_COSMETIC = (
        ("(c)", "&copy;"),
        ("(r)", "&reg;"),
        ("(tm)", "&trade;"),
    )

    for tag in ("i", "b", "u", "s", "sub", "sup", "h1", "h2", "h3", "h3", "h4", "h5", "h6"):
        parser.add_formatter(tag, render_tags)
    for align in ("left", "center", "right"):
        parser.add_simple_formatter(align, f'<code class="bbcode bbcode_{align}">%(value)s</code>')

    parser.add_simple_formatter("spoiler", '<span class="bbcode bbcode_spoiler">%(value)s</span>')
    parser.add_simple_formatter("url", '<a class="auto_link named_link">%(value)s</a>')
    parser.add_simple_formatter(
        "iconusername",
        f'<a class=iconusername href="/user/%(value)s">'
        f'<img alt="%(value)s" title="%(value)s" src="//a.furaffinity.net/{datetime.now():%Y%m%d}/%(value)s.gif">'
        f'%(value)s'
        f'</a>'
    )
    parser.add_simple_formatter(
        "usernameicon",
        f'<a class=iconusername href="/user/%(value)s">'
        f'<img alt="%(value)s" title="%(value)s" src="//a.furaffinity.net/{datetime.now():%Y%m%d}/%(value)s.gif">'
        f'</a>'
    )
    parser.add_simple_formatter("linkusername", '<a class=linkusername href="/user/%(value)s">%(value)s</a>')
    parser.add_simple_formatter("hr", "<hr>", standalone=True)

    parser.add_formatter("url", render_url)
    parser.add_formatter("color", render_color)
    parser.add_formatter("quote", render_quote)
    parser.add_formatter("tag", render_tag)

    bbcode = sub(r"-{5,}", "[hr]", bbcode)

    result_page: BeautifulSoup = parse_extra(parse_page(parser.format(bbcode)))
    return (result_page.select_one("html > body") or result_page).decode_contents()


def parse_mentions(tag: Tag) -> list[str]:
    mentions: list[str] = [username_url(m[1]) for a in tag.select("a")
                           if (m := match(mentions_regexp, get_attr(a, "href")))]
    return sorted(set([m for m in mentions if m]), key=mentions.index)


def parse_loggedin_user(page: BeautifulSoup) -> Optional[str]:
    return get_attr(avatar, "alt") if (avatar := page.select_one("img.loggedin_user_avatar")) else None


def parse_journal_section(section_tag: Tag) -> dict[str, Any]:
    id_: int = int(section_tag.attrs.get("id", "00000")[4:])
    tag_title: Optional[Tag] = section_tag.select_one("h2")
    tag_date: Optional[Tag] = section_tag.select_one("span.popup_date")
    tag_content: Optional[Tag] = section_tag.select_one("div.journal-body")
    tag_comments: Optional[Tag] = section_tag.select_one("div.section-footer > a > span")

    assert id_ != 0, _raise_exception(ParsingError("Missing ID"))
    assert tag_title is not None, _raise_exception(ParsingError("Missing title tag"))
    assert tag_date is not None, _raise_exception(ParsingError("Missing date tag"))
    assert tag_content is not None, _raise_exception(ParsingError("Missing content tag"))
    assert tag_comments is not None, _raise_exception(ParsingError("Missing comments tag"))

    # noinspection DuplicatedCode
    title: str = tag_title.text.strip()
    date: datetime = parse_date(
        get_attr(tag_date, "title").strip()
        if match(r"^[A-Za-z]+ \d+,.*$", get_attr(tag_date, "title"))
        else tag_date.text.strip()
    )
    content: str = clean_html(inner_html(tag_content))
    mentions: list[str] = parse_mentions(tag_content)
    comments: int = int(tag_comments.text.strip())

    return {
        "id": id_,
        "title": title,
        "date": date,
        "content": content,
        "mentions": mentions,
        "comments": comments,
    }


def parse_journal_page(journal_page: BeautifulSoup) -> dict[str, Any]:
    user_info: dict[str, str] = parse_user_folder(journal_page)
    tag_id: Optional[Tag] = journal_page.select_one("meta[property='og:url']")
    tag_title: Optional[Tag] = journal_page.select_one("h2.journal-title")
    tag_date: Optional[Tag] = journal_page.select_one("span.popup_date")
    tag_header: Optional[Tag] = journal_page.select_one("div.journal-header")
    tag_footer: Optional[Tag] = journal_page.select_one("div.journal-footer")
    tag_content: Optional[Tag] = journal_page.select_one("div.journal-content")
    tag_comments: Optional[Tag] = journal_page.select_one("div.section-footer > span")

    assert tag_id is not None, _raise_exception(ParsingError("Missing ID tag"))
    assert tag_title is not None, _raise_exception(ParsingError("Missing title tag"))
    assert tag_date is not None, _raise_exception(ParsingError("Missing date tag"))
    assert tag_content is not None, _raise_exception(ParsingError("Missing content tag"))
    assert tag_comments is not None, _raise_exception(ParsingError("Missing comments tag"))

    id_: int = int(tag_id.attrs.get("content", "0").strip("/").split("/")[-1])
    # noinspection DuplicatedCode
    title: str = tag_title.text.strip()
    date: datetime = parse_date(
        get_attr(tag_date, "title").strip()
        if match(r"^[A-Za-z]+ \d+,.*$", get_attr(tag_date, "title"))
        else tag_date.text.strip()
    )
    header: str = clean_html(inner_html(tag_header)) if tag_header else ""
    footer: str = clean_html(inner_html(tag_footer)) if tag_footer else ""
    content: str = clean_html(inner_html(tag_content))
    mentions: list[str] = parse_mentions(tag_content)
    comments: int = int(tag_comments.text.strip())

    assert id_ != 0, _raise_exception(ParsingError("Missing ID"))

    return {
        **user_info,
        "id": id_,
        "title": title,
        "date": date,
        "content": content,
        "header": header,
        "footer": footer,
        "mentions": mentions,
        "comments": comments,
    }


def parse_submission_figure(figure_tag: Tag) -> dict[str, Any]:
    id_: int = int(get_attr(figure_tag, "id")[4:])
    tag_title: Optional[Tag] = figure_tag.select_one("figcaption a[href^='/view/']")
    tag_author: Optional[Tag] = figure_tag.select_one("figcaption a[href^='/user/']")
    tag_thumbnail: Optional[Tag] = figure_tag.select_one("img")

    assert tag_title is not None, _raise_exception(ParsingError("Missing title tag"))
    assert tag_author is not None, _raise_exception(ParsingError("Missing author tag"))
    assert tag_thumbnail is not None, _raise_exception(ParsingError("Missing thumbnail tag"))

    title: str = get_attr(tag_title, "title")
    author: str = get_attr(tag_author, "title")
    rating: str = next(c for c in figure_tag["class"] if c.startswith("r-"))[2:]
    type_: str = next(c for c in figure_tag["class"] if c.startswith("t-"))[2:]
    thumbnail_url: str = "https:" + get_attr(tag_thumbnail, "src")

    return {
        "id": id_,
        "title": title,
        "author": author,
        "rating": rating,
        "type": type_,
        "thumbnail_url": thumbnail_url,
    }


def parse_submission_author(author_tag: Tag) -> dict[str, Any]:
    tag_author: Optional[Tag] = author_tag.select_one("div.submission-id-sub-container")

    assert tag_author is not None, _raise_exception(ParsingError("Missing author tag"))

    tag_author_name: Optional[Tag] = tag_author.select_one("a > strong")
    tag_author_icon: Optional[Tag] = author_tag.select_one("img.submission-user-icon")

    assert tag_author_name is not None, _raise_exception(ParsingError("Missing author name tag"))
    assert tag_author_icon is not None, _raise_exception(ParsingError("Missing author icon tag"))

    author_name: str = tag_author_name.text.strip()
    author_title: str = ([*filter(bool, [child.strip()
                                         for child in tag_author.children
                                         if isinstance(child, NavigableString)][3:])] or [""])[-1]
    author_title = author_title if tag_author.select_one('a[href$="/#tip"]') is None else sub(r"\|$", "", author_title)
    author_title = author_title.strip("\xA0 ")  # NBSP
    author_icon_url: str = "https:" + get_attr(tag_author_icon, "src")

    return {
        "author": author_name,
        "author_title": author_title,
        "author_icon_url": author_icon_url,
    }


def parse_submission_page(sub_page: BeautifulSoup) -> dict[str, Any]:
    tag_id: Optional[Tag] = sub_page.select_one("meta[property='og:url']")
    tag_sub_info: Optional[Tag] = sub_page.select_one("div.submission-id-sub-container")

    assert tag_sub_info is not None, _raise_exception(ParsingError("Missing info tag"))

    tag_title: Optional[Tag] = tag_sub_info.select_one("div.submission-title")
    tag_author: Optional[Tag] = sub_page.select_one("div.submission-id-container")
    tag_date: Optional[Tag] = sub_page.select_one("span.popup_date")
    tag_tags: list[Tag] = sub_page.select("section.tags-row a")
    tag_views: Optional[Tag] = sub_page.select_one("div.views span")
    tag_comment_count: Optional[Tag] = sub_page.select_one("section.stats-container div.comments span")
    tag_favorites: Optional[Tag] = sub_page.select_one("div.favorites span")
    tag_rating: Optional[Tag] = sub_page.select_one("div.rating span")
    tag_type: Optional[Tag] = sub_page.select_one("div#submission_page[class^='page-content-type']")
    tag_fav: Optional[Tag] = sub_page.select_one("div.fav > a")
    tag_info: Optional[Tag] = sub_page.select_one("section.info.text")
    tag_user_folders: list[Tag] = sub_page.select("section.folder-list-container > div > a")

    assert tag_info is not None, _raise_exception(ParsingError("Missing info tag"))

    tag_category1: Optional[Tag] = tag_info.select_one("span.category-name")
    tag_category2: Optional[Tag] = tag_info.select_one("span.type-name")
    tag_species: Optional[Tag] = (info_spans := tag_info.select("span"))[2]
    tag_gender: Optional[Tag] = info_spans[3]
    tag_description: Optional[Tag] = sub_page.select_one("div.submission-description")
    tag_folder: Optional[Tag] = sub_page.select_one("a.button[href^='/scraps/'],a.button[href^='/gallery/']")
    tag_file_url: Optional[Tag] = sub_page.select_one("div.download a")
    tag_thumbnail_url: Optional[Tag] = sub_page.select_one("img#submissionImg")
    tag_prev: Optional[Tag] = sub_page.select_one("div.submission-content div.favorite-nav a:nth-child(1)")
    tag_next: Optional[Tag] = sub_page.select_one("div.submission-content div.favorite-nav a:last-child")

    assert tag_id is not None, _raise_exception(ParsingError("Missing id tag"))
    assert tag_title is not None, _raise_exception(ParsingError("Missing title tag"))
    assert tag_author is not None, _raise_exception(ParsingError("Missing author tag"))
    assert tag_date is not None, _raise_exception(ParsingError("Missing date tag"))
    assert tag_views is not None, _raise_exception(ParsingError("Missing views tag"))
    assert tag_comment_count is not None, _raise_exception(ParsingError("Missing comment count tag"))
    assert tag_favorites is not None, _raise_exception(ParsingError("Missing favorites tag"))
    assert tag_rating is not None, _raise_exception(ParsingError("Missing rating tag"))
    assert tag_type is not None, _raise_exception(ParsingError("Missing type tag"))
    assert tag_fav is not None, _raise_exception(ParsingError("Missing fav tag"))
    assert tag_category1 is not None, _raise_exception(ParsingError("Missing category1 tag"))
    assert tag_category2 is not None, _raise_exception(ParsingError("Missing category2 tag"))
    assert tag_species is not None, _raise_exception(ParsingError("Missing species tag"))
    assert tag_gender is not None, _raise_exception(ParsingError("Missing gender tag"))
    assert tag_description is not None, _raise_exception(ParsingError("Missing description tag"))
    assert tag_folder is not None, _raise_exception(ParsingError("Missing folder tag"))
    assert tag_file_url is not None, _raise_exception(ParsingError("Missing file URL tag"))
    assert tag_prev is not None, _raise_exception(ParsingError("Missing prev tag"))
    assert tag_next is not None, _raise_exception(ParsingError("Missing next tag"))

    tag_footer: Optional[Tag] = tag_description.select_one("div.submission-footer")

    id_: int = int(get_attr(tag_id, "content").strip("/").split("/")[-1])
    title: str = tag_title.text.strip()
    date: datetime = parse_date(
        get_attr(tag_date, "title").strip()
        if match(r"^[A-Za-z]+ \d+,.*$", get_attr(tag_date, "title"))
        else tag_date.text.strip()
    )
    tags: list[str] = [t.text.strip() for t in tag_tags]
    category: str = tag_category1.text.strip() + " / " + tag_category2.text.strip()
    species: str = tag_species.text.strip()
    gender: str = tag_gender.text.strip()
    rating: str = tag_rating.text.strip()
    views: int = int(tag_views.text.strip())
    comment_count: int = int(tag_comment_count.text.strip())
    favorites: int = int(tag_favorites.text.strip())
    type_: str = tag_type["class"][0][18:]
    footer: str = ""
    if tag_footer:
        if tag_footer_hr := tag_footer.select_one("hr"):
            tag_footer_hr.decompose()
        footer = clean_html(inner_html(tag_footer))
        tag_footer.decompose()
    description: str = clean_html(inner_html(tag_description))
    mentions: list[str] = parse_mentions(tag_description)
    folder: str = m.group(1).lower() if (m := match(r"^/(scraps|gallery)/.*$", get_attr(tag_folder, "href"))) else ""
    file_url: str = "https:" + get_attr(tag_file_url, "href")
    thumbnail_url: str = ("https:" + get_attr(tag_thumbnail_url, "data-preview-src")) if tag_thumbnail_url else ""
    prev_sub: Optional[int] = int(
        get_attr(tag_prev, "href").split("/")[-2]) if tag_prev and tag_prev.text.lower() == "prev" else None
    next_sub: Optional[int] = int(
        get_attr(tag_next, "href").split("/")[-2]) if tag_next and tag_next.text.lower() == "next" else None
    fav_link: Optional[str] = f"{root}{href}" if (href := get_attr(tag_fav, "href")).startswith("/fav/") else None
    unfav_link: Optional[str] = f"{root}{href}" if (href := get_attr(tag_fav, "href")).startswith("/unfav/") else None
    user_folders: list[tuple[str, str, str]] = []
    for a in tag_user_folders:
        tag_folder_name: Optional[Tag] = a.select_one("span")
        tag_folder_group: Optional[Tag] = a.select_one("strong")
        assert tag_folder_name is not None, _raise_exception(ParsingError("Missing folder name tag"))
        user_folders.append((
            tag_folder_name.text.strip(),
            (root + href) if (href := a.attrs.get("href", "")) else "",
            tag_folder_group.text.strip() if tag_folder_group else ""
        ))

    return {
        "id": id_,
        "title": title,
        **parse_submission_author(tag_author),
        "date": date,
        "tags": tags,
        "category": category,
        "species": species,
        "gender": gender,
        "rating": rating,
        "views": views,
        "comment_count": comment_count,
        "favorites": favorites,
        "type": type_,
        "footer": footer,
        "description": description,
        "mentions": mentions,
        "folder": folder,
        "user_folders": user_folders,
        "file_url": file_url,
        "thumbnail_url": thumbnail_url,
        "prev": prev_sub,
        "next": next_sub,
        "fav_link": fav_link,
        "unfav_link": unfav_link,
    }


def parse_user_page(user_page: BeautifulSoup) -> dict[str, Any]:
    tag_status: Optional[Tag] = user_page.select_one("div.username h2")
    tag_profile: Optional[Tag] = user_page.select_one("div.userpage-profile")
    tag_title_join_date: Optional[Tag] = user_page.select_one("div.userpage-flex-item.username > span")
    tag_stats: Optional[Tag] = user_page.select_one("div.userpage-section-right div.table")
    tag_watchlist_to: Optional[Tag] = user_page.select_one("a[href*='watchlist/to']")
    tag_watchlist_by: Optional[Tag] = user_page.select_one("a[href*='watchlist/by']")
    tag_infos: list[Tag] = user_page.select("div#userpage-contact-item div.table-row")
    tag_contacts: list[Tag] = user_page.select("div#userpage-contact div.user-contact-user-info")
    tag_user_icon_url: Optional[Tag] = user_page.select_one("img.user-nav-avatar")
    tag_user_nav_controls: Optional[Tag] = user_page.select_one("div.user-nav-controls")

    assert tag_status is not None, _raise_exception(ParsingError("Missing name tag"))
    assert tag_profile is not None, _raise_exception(ParsingError("Missing profile tag"))
    assert tag_title_join_date is not None, _raise_exception(ParsingError("Missing join date tag"))
    assert tag_stats is not None, _raise_exception(ParsingError("Missing stats tag"))
    assert tag_watchlist_to is not None, _raise_exception(ParsingError("Missing watchlist to tag"))
    assert tag_watchlist_by is not None, _raise_exception(ParsingError("Missing watchlist by tag"))
    assert tag_user_icon_url is not None, _raise_exception(ParsingError("Missing user icon URL tag"))
    assert tag_user_nav_controls is not None, _raise_exception(ParsingError("Missing user nav controls tag"))

    tag_watch: Optional[Tag] = tag_user_nav_controls.select_one("a[href^='/watch/'], a[href^='/unwatch/']")
    tag_block: Optional[Tag] = tag_user_nav_controls.select_one("a[href^='/block/'], a[href^='/unblock/']")

    status: str = (u := tag_status.text.strip())[0]
    name: str = u[1:]
    title: str = ttd[0].strip() if len(ttd := tag_title_join_date.text.rsplit("|", 1)) > 1 else ""
    join_date: datetime = parse_date(ttd[-1].strip().split(":", 1)[1])
    profile: str = clean_html(inner_html(tag_profile))
    stats: tuple[int, ...] = (
        *map(lambda s: int(s.split(":")[1]), filter(bool, map(str.strip, tag_stats.text.split("\n")))),
        int(m[1]) if (m := search(r"(\d+)", tag_watchlist_to.text)) else 0,
        int(m[1]) if (m := search(r"(\d+)", tag_watchlist_by.text)) else 0,
    )

    tag_key: Optional[Tag]
    info: dict[str, str] = {}
    contacts: dict[str, str] = {}
    for tb in tag_infos:
        if (tag_key := tb.select_one("div")) is None:
            continue
        elif "profile-empty" in tb.attrs.get("class", []):
            continue
        elif not (val := [*filter(bool, [c.strip() for c in tb.children if isinstance(c, NavigableString)])][-1:]):
            continue
        info[tag_key.text.strip()] = val[0]
    for pc in tag_contacts:
        if (tag_key := pc.select_one("span")) is None:
            continue
        contacts[tag_key.text.strip()] = get_attr(a, "href") if (a := pc.select_one("a")) else \
            [*filter(bool, map(str.strip, pc.text.split("\n")))][-1]
    user_icon_url: str = "https:" + get_attr(tag_user_icon_url, "src")
    tag_watch_href: str = get_attr(tag_watch, "href") if tag_watch else ""
    watch: Optional[str] = f"{root}{tag_watch_href}" if tag_watch_href.startswith("/watch/") else None
    unwatch: Optional[str] = f"{root}{tag_watch_href}" if tag_watch_href.startswith("/unwatch/") else None
    tag_block_href: str = get_attr(tag_block, "href") if tag_block else ""
    block: Optional[str] = f"{root}{tag_block_href}" if tag_block_href.startswith("/block/") else None
    unblock: Optional[str] = f"{root}{tag_block_href}" if tag_block_href.startswith("/unblock/") else None

    return {
        "name": name,
        "status": status,
        "title": title,
        "join_date": join_date,
        "profile": profile,
        "stats": stats,
        "info": info,
        "contacts": contacts,
        "user_icon_url": user_icon_url,
        "watch": watch,
        "unwatch": unwatch,
        "block": block,
        "unblock": unblock,
    }


def parse_comment_tag(tag: Tag) -> dict:
    tag_id: Optional[Tag] = tag.select_one("a")
    tag_username: Optional[Tag] = tag.select_one("div.header strong.comment_username > h3")
    tag_user_icon: Optional[Tag] = tag.select_one("div.header img.comment_useravatar")
    tag_user_title: Optional[Tag] = tag.select_one("div.header div.name div.cell span.hideonmobile")
    tag_body: Optional[Tag] = tag.select_one("div.base > div.body")
    tag_parent_link: Optional[Tag] = tag.select_one("a.comment-parent")
    tag_edited: Optional[Tag] = tag.select_one("img.edited")

    assert tag_id is not None, _raise_exception(ParsingError("Missing link tag"))
    assert tag_body is not None, _raise_exception(ParsingError("Missing body tag"))

    attr_id: Optional[str] = tag_id.attrs.get("id")

    assert attr_id is not None, _raise_exception(ParsingError("Missing id attribute"))

    comment_id: int = int(attr_id.removeprefix("cid:"))
    comment_text: str = clean_html(inner_html(tag_body))

    if tag_username is None:
        return {
            "id": comment_id,
            "user_name": "",
            "user_title": "",
            "user_icon_url": "",
            "timestamp": 0,
            "text": comment_text,
            "parent": None,
            "edited": tag_edited is not None,
            "hidden": True,
        }

    assert tag_username is not None, _raise_exception(ParsingError("Missing user name tag"))
    assert tag_user_icon is not None, _raise_exception(ParsingError("Missing user icon tag"))
    assert tag_user_title is not None, _raise_exception(ParsingError("Missing user title tag"))

    attr_timestamp: Optional[str] = tag.attrs.get("data-timestamp")
    attr_user_icon: Optional[str] = tag_user_icon.attrs.get("src")
    attr_parent_href: Optional[str] = tag_parent_link.attrs.get("href") if tag_parent_link is not None else None

    assert attr_timestamp is not None, _raise_exception(ParsingError("Missing timestamp attribute"))
    assert attr_user_icon is not None, _raise_exception(ParsingError("Missing user icon src attribute"))

    parent_id: Optional[int] = int(attr_parent_href.removeprefix("#cid:")) if attr_parent_href else None

    return {
        "id": comment_id,
        "user_name": tag_username.text.strip(),
        "user_title": tag_user_title.text.strip(),
        "user_icon_url": "https:" + attr_user_icon,
        "timestamp": int(attr_timestamp),
        "text": comment_text,
        "parent": parent_id,
        "edited": tag_edited is not None,
        "hidden": False,
    }


def parse_comments(page: BeautifulSoup) -> list[Tag]:
    return page.select("div.comment_container")


def parse_user_tag(user_tag: Tag) -> dict[str, Any]:
    status: str = (u := [*filter(bool, map(str.strip, user_tag.text.split("\n")))])[0][0]
    name: str = u[0][1:]
    title: str = ttd[0].strip() if len(ttd := u[1].rsplit("|", 1)) > 1 else ""
    join_date: datetime = parse_date(ttd[-1].strip().split(":", 1)[1])

    return {
        "user_name": name,
        "user_status": status,
        "user_title": title,
        "user_join_date": join_date,
    }


def parse_user_folder(folder_page: BeautifulSoup) -> dict[str, Any]:
    tag_username: Optional[Tag] = folder_page.select_one("div.userpage-flex-item.username")
    tag_user_icon: Optional[Tag] = folder_page.select_one("img.user-nav-avatar")
    assert tag_username is not None, _raise_exception(ParsingError("Missing username tag"))
    assert tag_user_icon is not None, _raise_exception(ParsingError("Missing user icon tag"))
    return {
        **parse_user_tag(tag_username),
        "user_icon_url": "https:" + get_attr(tag_user_icon, "src"),
    }


def parse_submission_figures(figures_page: BeautifulSoup) -> list[Tag]:
    return figures_page.select("figure[id^='sid-']")


def parse_user_submissions(submissions_page: BeautifulSoup) -> dict[str, Any]:
    user_info: dict[str, str] = parse_user_folder(submissions_page)
    last_page: bool = not any(b.text.lower() == "next" for b in submissions_page.select("form button.button"))

    return {
        **user_info,
        "figures": parse_submission_figures(submissions_page),
        "last_page": last_page,
    }


def parse_user_favorites(favorites_page: BeautifulSoup) -> dict[str, Any]:
    parsed_submissions = parse_user_submissions(favorites_page)
    tag_next_page: Optional[Tag] = favorites_page.select_one("a[class~=button][class~=standard][class~=right]")
    next_page: str = get_attr(tag_next_page, "href").split("/", 3)[-1] if tag_next_page else ""

    return {
        **parsed_submissions,
        "next_page": next_page,
    }


def parse_user_journals(journals_page: BeautifulSoup) -> dict[str, Any]:
    user_info: dict[str, str] = parse_user_folder(journals_page)
    sections: list[Tag] = journals_page.select("section[id^='jid:']")
    next_page_tag: Optional[Tag] = journals_page.select_one("div.mini-nav > div.mini-nav-cell:first-child > a.button")

    return {
        **user_info,
        "sections": sections,
        "last_page": next_page_tag is None,
    }


def parse_watchlist(watch_page: BeautifulSoup) -> tuple[list[tuple[str, str]], int]:
    tags_users: list[Tag] = watch_page.select("div.watch-list-items")
    tag_next: Optional[Tag] = watch_page.select_one("section div.floatright form[method=get]")
    match_next: Optional[Match] = watchlist_next_regexp.match(get_attr(tag_next, "action")) if tag_next else None
    return (
        [((u := t.text.strip().replace(" ", ""))[0], u[1:]) for t in tags_users],
        int(match_next[1]) if match_next else 0,
    )



from collections import namedtuple
from datetime import datetime
from typing import Optional
from typing import Union

from .connection import join_url
from .connection import root
from .exceptions import _raise_exception
from .parse import BeautifulSoup
from .parse import check_page_raise
from .parse import html_to_bbcode
from .parse import parse_comments
from .parse import parse_journal_page
from .parse import parse_journal_section
from .parse import Tag
from .user import UserPartial


class JournalStats(namedtuple("JournalStats", ["comments"])):
    """
    This object contains the journal's statistics:
    * comments
    """


class JournalBase:
    def __init__(self):
        self.id: int = 0
        self.title: str = ""
        self.date: datetime = datetime.fromtimestamp(0)
        self.author: UserPartial = UserPartial()
        self.stats: JournalStats = JournalStats(0)
        self.content: str = ""
        self.mentions: list[str] = []

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other) -> bool:
        if isinstance(other, JournalBase):
            return other.id == self.id
        elif isinstance(other, int):
            return other == self.id
        return False

    def __gt__(self, other) -> bool:
        if isinstance(other, JournalBase):
            return self.id > other.id
        elif isinstance(other, int):
            return self.id > other
        return False

    def __ge__(self, other) -> bool:
        if isinstance(other, JournalBase):
            return self.id >= other.id
        elif isinstance(other, int):
            return self.id >= other
        return False

    def __lt__(self, other) -> bool:
        if isinstance(other, JournalBase):
            return self.id < other.id
        elif isinstance(other, int):
            return self.id < other
        return False

    def __le__(self, other) -> bool:
        if isinstance(other, JournalBase):
            return self.id <= other.id
        elif isinstance(other, int):
            return self.id <= other
        return False

    def __iter__(self):
        yield "id", self.id
        yield "title", self.title
        yield "date", self.date
        yield "author", dict(self.author)
        yield "stats", self.stats._asdict()
        yield "content", self.content
        yield "mentions", self.mentions

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"{self.id} {self.author} {self.title}"

    @property
    def content_bbcode(self) -> str:
        """
        The journal content formatted to BBCode

        :return: BBCode content
        """
        return html_to_bbcode(self.content)

    @property
    def url(self) -> str:
        """
        Compose the full URL to the journal.

        :return: The URL to the journal.
        """
        return join_url(root, "journal", self.id)


class JournalPartial(JournalBase):
    """
    Contains partial journal information gathered from journals pages.
    """

    def __init__(self, journal_tag: Tag = None):
        """
        :param journal_tag: The tag from which to parse the journal.
        """
        assert journal_tag is None or isinstance(journal_tag, Tag), \
            _raise_exception(TypeError(f"journal_item must be {None} or {Tag.__name__}"))
        self.journal_tag: Optional[Tag] = journal_tag

        super(JournalPartial, self).__init__()

        self.parse()

    def parse(self, journal_tag: Union[Tag, BeautifulSoup] = None):
        """
        Parse a journal tag, overrides any information already present in the object.

        :param journal_tag: The tag from which to parse the journal.
        """
        assert journal_tag is None or isinstance(journal_tag, BeautifulSoup), \
            _raise_exception(TypeError(f"journal_item must be {None} or {BeautifulSoup.__name__}"))

        self.journal_tag = journal_tag or self.journal_tag
        if self.journal_tag is None:
            return

        parsed: dict = parse_journal_section(self.journal_tag)

        # noinspection DuplicatedCode
        self.id = parsed["id"]
        self.title = parsed["title"]
        self.author.name = parsed.get("user_name", "")
        self.author.status = parsed.get("user_status", "")
        self.author.title = parsed.get("user_title", "")
        self.author.join_date = parsed.get("user_join_date", "")
        self.author.user_icon_url = parsed.get("user_icon_url", "")
        self.stats = JournalStats(parsed["comments"])
        self.date = parsed["date"]
        self.content = parsed["content"]
        self.mentions = parsed["mentions"]


class Journal(JournalBase):
    """
    Contains complete journal information gathered from journal pages, including comments.
    """

    def __init__(self, journal_page: BeautifulSoup = None):
        """
        :param journal_page: The page from which to parse the journal.
        """
        assert journal_page is None or isinstance(journal_page, BeautifulSoup), \
            _raise_exception(TypeError(f"journal_item must be {None} or {BeautifulSoup.__name__}"))
        self.journal_page: Optional[BeautifulSoup] = journal_page

        super(Journal, self).__init__()

        self.header: str = ""
        self.footer: str = ""
        from .comment import Comment
        self.comments: list[Comment] = []

        self.parse()

    def __iter__(self):
        for k, v in super(Journal, self).__iter__():
            yield k, v
        yield "header", self.header
        yield "footer", self.footer
        from .comment import _remove_recursion
        yield "comments", [dict(_remove_recursion(c)) for c in self.comments]

    @property
    def header_bbcode(self) -> str:
        """
        The journal header formatted to BBCode

        :return: BBCode header
        """
        return html_to_bbcode(self.header)

    @property
    def footer_bbcode(self) -> str:
        """
        The journal footer formatted to BBCode

        :return: BBCode footer
        """
        return html_to_bbcode(self.footer)

    def parse(self, journal_page: Union[Tag, BeautifulSoup] = None):
        """
        Parse a journal page, overrides any information already present in the object.

        :param journal_page: The page from which to parse the journal.
        """
        assert journal_page is None or isinstance(journal_page, BeautifulSoup), \
            _raise_exception(TypeError(f"journal_item must be {None} or {BeautifulSoup.__name__}"))

        self.journal_page = journal_page or self.journal_page
        if self.journal_page is None:
            return

        check_page_raise(self.journal_page)

        parsed: dict = parse_journal_page(self.journal_page)

        # noinspection DuplicatedCode
        self.id = parsed["id"]
        self.title = parsed["title"]
        self.author.name = parsed.get("user_name", "")
        self.author.status = parsed.get("user_status", "")
        self.author.title = parsed.get("user_title", "")
        self.author.join_date = parsed.get("user_join_date", "")
        self.author.user_icon_url = parsed.get("user_icon_url", "")
        self.stats = JournalStats(parsed["comments"])
        self.date = parsed["date"]
        self.content = parsed["content"]
        self.header = parsed["header"]
        self.footer = parsed["footer"]
        self.mentions = parsed["mentions"]
        from .comment import sort_comments, Comment
        self.comments = sort_comments([Comment(t, self) for t in parse_comments(self.journal_page)])



class DisallowedPath(Exception):
    """
    The path is not allowed by the robots.txt.
    """


class Unauthorized(Exception):
    """
    The user is not logged-in.
    """


class ParsingError(Exception):
    """
    An error occurred while parsing the page.
    """


class NonePage(ParsingError):
    """
    The parsed page is None.
    """


class NoTitle(ParsingError):
    """
    The parsed paged is missing a title.
    """


class NotFound(ParsingError):
    """
    The resource could not be found.
    """


class DisabledAccount(ParsingError):
    """
    The resource belongs to a disabled account.
    """


class ServerError(ParsingError):
    """
    The page contains a server error notice.
    """


class NoticeMessage(ParsingError):
    """
    A notice of unknown type was found in the page.
    """


def _raise_exception(err: BaseException):
    raise err



from collections import namedtuple
from datetime import datetime
from typing import Optional

from .connection import join_url
from .connection import root
from .exceptions import _raise_exception
from .parse import BeautifulSoup
from .parse import Tag
from .parse import check_page_raise
from .parse import html_to_bbcode
from .parse import parse_comments
from .parse import parse_submission_figure
from .parse import parse_submission_page
from .user import UserPartial


class SubmissionStats(namedtuple("SubmissionStats", ["views", "comments", "favorites"])):
    """
    This object contains the submission's statistics:
    * views
    * comments
    * favorites
    """


class SubmissionUserFolder(namedtuple("SubmissionUserFolder", ["name", "url", "group"])):
    """
    This object contains a submission's folder details:
    * name: str the name of the folder
    * url: str the URL to the folder
    * group: str the group the folder belongs to
    """


class SubmissionBase:
    """
    Base class for the submission objects.
    """

    def __init__(self):
        self.id: int = 0
        self.title: str = ""
        self.author: UserPartial = UserPartial()

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other) -> bool:
        if isinstance(other, SubmissionBase):
            return other.id == self.id
        elif isinstance(other, int):
            return other == self.id
        return False

    def __gt__(self, other) -> bool:
        if isinstance(other, SubmissionBase):
            return self.id > other.id
        elif isinstance(other, int):
            return self.id > other
        return False

    def __ge__(self, other) -> bool:
        if isinstance(other, SubmissionBase):
            return self.id >= other.id
        elif isinstance(other, int):
            return self.id >= other
        return False

    def __lt__(self, other) -> bool:
        if isinstance(other, SubmissionBase):
            return self.id < other.id
        elif isinstance(other, int):
            return self.id < other
        return False

    def __le__(self, other) -> bool:
        if isinstance(other, SubmissionBase):
            return self.id <= other.id
        elif isinstance(other, int):
            return self.id <= other
        return False

    def __iter__(self):
        yield "id", self.id
        yield "title", self.title
        yield "author", dict(self.author)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"{self.id} {self.author} {self.title}"

    @property
    def url(self):
        """
        Compose the full URL to the submission.

        :return: The URL to the submission.
        """
        return join_url(root, "view", self.id)


class SubmissionPartial(SubmissionBase):
    """
    Contains partial submission information gathered from submissions pages (gallery, scraps, etc.).
    """

    def __init__(self, submission_figure: Tag = None):
        """
        :param submission_figure: The figure tag from which to parse the submission information.
        """
        assert submission_figure is None or isinstance(submission_figure, Tag), \
            _raise_exception(TypeError(f"submission_figure must be {None} or {BeautifulSoup.__name__}"))

        super().__init__()

        self.submission_figure: Optional[Tag] = submission_figure
        self.rating: str = ""
        self.type: str = ""
        self.thumbnail_url: str = ""

        self.parse()

    def __iter__(self):
        yield "id", self.id
        yield "title", self.title
        yield "author", dict(self.author)
        yield "rating", self.rating
        yield "type", self.type
        yield "thumbnail_url", self.thumbnail_url

    def parse(self, submission_figure: Tag = None):
        """
        Parse a submission figure Tag, overrides any information already present in the object.

        :param submission_figure: The optional figure tag from which to parse the submission.
        """
        assert submission_figure is None or isinstance(submission_figure, Tag), \
            _raise_exception(TypeError(f"submission_figure must be {None} or {BeautifulSoup.__name__}"))

        self.submission_figure = submission_figure or self.submission_figure
        if self.submission_figure is None:
            return

        parsed: dict = parse_submission_figure(self.submission_figure)

        self.id = parsed["id"]
        self.title = parsed["title"]
        self.author.name = parsed["author"]
        self.rating = parsed["rating"]
        self.type = parsed["type"]
        self.thumbnail_url = parsed["thumbnail_url"]


class Submission(SubmissionBase):
    """
    Contains complete submission information gathered from submission pages, including comments.
    """

    def __init__(self, submission_page: BeautifulSoup = None):
        """
        :param submission_page: The page from which to parse the submission information.
        """
        assert submission_page is None or isinstance(submission_page, BeautifulSoup), \
            _raise_exception(TypeError(f"submission_page must be {None} or {BeautifulSoup.__name__}"))

        super().__init__()

        self.submission_page: Optional[BeautifulSoup] = submission_page
        self.date: datetime = datetime.fromtimestamp(0)
        self.tags: list[str] = []
        self.category: str = ""
        self.species: str = ""
        self.gender: str = ""
        self.rating: str = ""
        self.stats: SubmissionStats = SubmissionStats(0, 0, 0)
        self.type: str = ""
        self.description: str = ""
        self.footer: str = ""
        self.mentions: list[str] = []
        self.folder: str = ""
        self.user_folders: list[SubmissionUserFolder] = []
        self.file_url: str = ""
        self.thumbnail_url: str = ""
        self.prev: Optional[int] = None
        self.next: Optional[int] = None
        self.favorite: bool = False
        self.favorite_toggle_link: str = ""
        from .comment import Comment
        self.comments: list[Comment] = []

        self.parse()

    def __iter__(self):
        yield "id", self.id
        yield "title", self.title
        yield "author", dict(self.author)
        yield "date", self.date
        yield "tags", self.tags
        yield "category", self.category
        yield "species", self.species
        yield "gender", self.gender
        yield "rating", self.rating
        yield "stats", self.stats._asdict()
        yield "type", self.type
        yield "description", self.description
        yield "footer", self.footer
        yield "mentions", self.mentions
        yield "folder", self.folder
        yield "user_folders", [f._asdict() for f in self.user_folders]
        yield "file_url", self.file_url
        yield "thumbnail_url", self.thumbnail_url
        yield "prev", self.prev
        yield "next", self.next
        yield "favorite", self.favorite
        yield "favorite_toggle_link", self.favorite_toggle_link
        from .comment import _remove_recursion
        yield "comments", [dict(_remove_recursion(c)) for c in self.comments]

    @property
    def description_bbcode(self) -> str:
        """
        The submission description formatted to BBCode

        :return: BBCode description
        """
        return html_to_bbcode(self.description)

    @property
    def footer_bbcode(self) -> str:
        """
        The submission footer formatted to BBCode

        :return: BBCode footer
        """
        return html_to_bbcode(self.footer)

    def parse(self, submission_page: BeautifulSoup = None):
        """
        Parse a submission page, overrides any information already present in the object.

        :param submission_page: The optional page from which to parse the submission.
        """
        assert submission_page is None or isinstance(submission_page, BeautifulSoup), \
            _raise_exception(TypeError(f"submission_page must be {None} or {BeautifulSoup.__name__}"))

        self.submission_page = submission_page or self.submission_page
        if self.submission_page is None:
            return

        check_page_raise(self.submission_page)

        parsed: dict = parse_submission_page(self.submission_page)

        self.id = parsed["id"]
        self.title = parsed["title"]
        self.author.name = parsed["author"]
        self.author.title = parsed["author_title"]
        self.author.user_icon_url = parsed["author_icon_url"]
        self.date = parsed["date"]
        self.tags = parsed["tags"]
        self.category = parsed["category"]
        self.species = parsed["species"]
        self.gender = parsed["gender"]
        self.rating = parsed["rating"]
        self.stats = SubmissionStats(parsed["views"], parsed["comment_count"], parsed["favorites"])
        self.type = parsed["type"]
        self.description = parsed["description"]
        self.footer = parsed["footer"]
        self.mentions = parsed["mentions"]
        self.folder = parsed["folder"]
        self.user_folders = [SubmissionUserFolder(*f) for f in parsed["user_folders"]]
        self.file_url = parsed["file_url"]
        self.thumbnail_url = parsed["thumbnail_url"]
        self.prev = parsed["prev"]
        self.next = parsed["next"]
        self.favorite = parsed["unfav_link"] is not None
        self.favorite_toggle_link = parsed["fav_link"] or parsed["unfav_link"]
        from .comment import sort_comments, Comment
        self.comments = sort_comments([Comment(t, self) for t in parse_comments(self.submission_page)])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['faapi']

package_data = \
{'': ['*']}

install_requires = \
['bbcode>=1.1.0,<2.0.0',
 'beautifulsoup4>=4.11.1,<5.0.0',
 'cfscrape>=2.1.1,<3.0.0',
 'htmlmin>=0.1.12,<0.2.0',
 'lxml>=4.9.1,<5.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'requests>=2.28.1,<3.0.0']

setup_kwargs = {
    'name': 'faapi',
    'version': '3.9.0',
    'description': 'Python module to implement API-like functionality for the FurAffinity.net website.',
    'long_description': '<div align="center">\n\n<img alt="logo" width="400" src="https://raw.githubusercontent.com/FurryCoders/Logos/main/logos/faapi-transparent.png">\n\n# Fur Affinity API\n\nPython library to implement API-like functionality for the [Fur Affinity](https://furaffinity.net) website.\n\n[![](https://img.shields.io/pypi/v/faapi?logo=pypi)](https://pypi.org/project/faapi/)\n[![](https://img.shields.io/pypi/pyversions/faapi?logo=Python)](https://www.python.org)\n\n[![](https://img.shields.io/github/v/tag/FurryCoders/faapi?label=github&sort=date&logo=github&color=blue)](https://github.com/FurryCoders/faapi)\n[![](https://img.shields.io/github/issues/FurryCoders/faapi?logo=github&color=blue)](https://github.com/FurryCoders/FAAPI/issues)\n[![](https://img.shields.io/github/workflow/status/FurryCoders/FAAPI/Test?label=Test&logo=githubactions)](https://github.com/FurryCoders/FAAPI/actions/workflows/test.yml)\n\n</div>\n\n## Requirements\n\nPython 3.9+ is necessary to run this\nlibrary. [<img src="https://python-poetry.org/images/logo-origami.svg" height="12"> Poetry](https://python-poetry.org)\nis used for packaging and dependency management.\n\n## Usage\n\nThe API comprises a main class `FAAPI`, two submission classes `Submission` and `SubmissionPartial`, a journal\nclass `Journal`, and a user class `User`.\n\nOnce `FAAPI` is initialized, its methods can be used to crawl FA and return parsed objects.\n\n```python\nfrom requests.cookies import RequestsCookieJar\nimport faapi\nimport orjson\n\ncookies = RequestsCookieJar()\ncookies.set("a", "38565475-3421-3f21-7f63-3d341339737")\ncookies.set("b", "356f5962-5a60-0922-1c11-65003b70308")\n\napi = faapi.FAAPI(cookies)\nsub, sub_file = api.submission(12345678, get_file=True)\n\nprint(sub.id, sub.title, sub.author, f"{len(sub_file) / 1024:02f}KiB")\n\nwith open(f"{sub.id}.json", "wb") as f:\n    f.write(orjson.dumps(dict(sub)))\n\nwith open(sub.file_url.split("/")[-1], "wb") as f:\n    f.write(sub_file)\n\ngallery, _ = api.gallery("user_name", 1)\nwith open("user_name-gallery.json", "wb") as f:\n    f.write(orjson.dumps(list(map(dict, gallery))))\n```\n\n### robots.txt\n\nAt init, the `FAAPI` object downloads the [robots.txt](https://www.furaffinity.net/robots.txt) file from FA to determine\nthe `Crawl-delay` and `disallow` values set therein. If not set in the robots.txt file, a crawl delay value of 1 second\nis used.\n\nTo respect this value, the default behaviour of the `FAAPI` object is to wait when a get request is made if the last\nrequest was performed more recently then the crawl delay value.\n\nSee under [FAAPI](#faapi) for more details on this behaviour.\n\nFurthermore, any get operation that points to a disallowed path from robots.txt will raise an exception. This check\nshould not be circumvented, and the developer of this library does not take responsibility for violations of the TOS of\nFur Affinity.\n\n### Cookies\n\nTo access protected pages, cookies from an active session are needed. These cookies can be given to the FAAPI object as\na list of dictionaries - each containing a `name` and a `value` field -, or as a `http.cookiejar.CookieJar`\nobject (`requests.cookies.RequestsCookieJar` and other objects inheriting from `CookieJar` are also supported). The\ncookies list should look like the following example:\n\n```python\ncookies = [\n    {"name": "a", "value": "38565475-3421-3f21-7f63-3d3413397537"},\n    {"name": "b", "value": "356f5962-5a60-0922-1c11-65003b703038"},\n]\n```\n\n```python\nfrom requests.cookies import RequestsCookieJar\n\ncookies = RequestsCookieJar()\ncookies.set("a", "38565475-3421-3f21-7f63-3d3413397537")\ncookies.set("b", "356f5962-5a60-0922-1c11-65003b703038")\n```\n\nTo access session cookies, consult the manual of the browser used to log in.\n\n*Note:* it is important to not logout of the session the cookies belong to, otherwise they will no longer work.<br/>\n*Note:* as of April 2022 only cookies `a` and `b` are needed.\n\n### User Agent\n\n`FAAPI` attaches a `User-Agent` header to every request. The user agent string is generated at startup in the following\nformat: `faapi/{library version} Python/{python version} {system name}/{system release}`.\n\n## Objects\n\n### FAAPI\n\nThis is the main object that handles all the calls to scrape pages and get submissions.\n\nIt holds 6 different fields:\n\n* `session: CloudflareScraper` `cfscrape` session used for get requests\n* `robots: urllib.robotparser.RobotFileParser` robots.txt handler\n* `user_agent: str` user agent used by the session (property, cannot be set)\n* `crawl_delay: float` crawl delay from robots.txt (property, cannot be set)\n* `last_get: float` time of last get (UNIX time)\n* `raise_for_unauthorized: bool = True` if set to `True`, raises an exception if a request is made and the resulting\n  page is not from a login session\n* `timeout: int | None = None` requests timeout in seconds for both page requests (e.g. submissions) and files\n\n#### Init\n\n`__init__(cookies: list[dict[str, str]] | CookieJar)`\n\nThe class init has a single argument for the cookies. Cookies must be in the format mentioned above\nin [#Cookies](#cookies).\n\n#### Methods & Properties\n\n* `load_cookies(cookies: list[dict[str, str]] | CookieJar)`<br/>\n  Load new cookies and create a new session.<br/>\n  *Note:* This method removes any cookies currently in use, to update/add single cookies access them from the session\n  object.\n* `handle_delay()`<br/>\n  Handles the crawl delay as set in the robots.txt\n* `check_path(path: str, *, raise_for_disallowed: bool = False) -> bool`<br/>\n  Checks whether a given path is allowed by the robots.txt. If `raise_for_disallowed` is set to `True`\n  a `DisallowedPath` exception is raised on non-allowed paths.\n* `connection_status -> bool`<br/>\n  Returns the status of the connection.\n* `login_status -> bool`<br/>\n  Returns the login status.\n* `get(path: str, **params) -> requests.Response`<br/>\n  This returns a response object containing the result of the get operation on the given URL with the\n  optional `**params` added to it (url provided is considered as path from \'https://www.furaffinity.net/\').\n* `get_parsed(path: str, *, skip_page_check: bool = False, skip_auth_check: bool = False, **params) -> bs4.BeautifulSoup`<br/>\n  Similar to `get()` but returns the parsed HTML from the normal get operation. If the GET request encountered an error,\n  an `HTTPError` exception is raised. If `skip_page_check` is set to `True`, the parsed page is not checked for errors (\n  e.g. non-existing submission). If `skip_auth_check` is set to `True`, the page is not checked for login status.\n* `me() -> User | None`<br/>\n  Returns the logged-in user as a `User` object if the cookies are from a login session.\n* `frontpage() -> list[SubmissionPartial]`<br/>\n  Fetch the latest submissions from Fur Affinity\'s front page.\n* `submission(submission_id: int, get_file: bool = False, *, chunk_size: int = None) -> tuple[Submission, bytes | None]`<br/>\n  Given a submission ID, it returns a `Submission` object containing the various metadata of the submission itself and\n  a `bytes` object with the submission file if `get_file` is passed as `True`. The optional `chunk_size` argument is\n  used for the request; if left to `None` or set to 0 the download is performed directly without streaming.<br/>\n  *Note:* the author `UserPartial` object of the submission does not contain the `join_date` field as it does not appear\n  on submission pages.\n* `submission_file(submission: Submission, *, chunk_size: int = None) -> bytes`<br/>\n  Given a submission object, it downloads its file and returns it as a `bytes` object. The optional `chunk_size`\n  argument is used for the request; if left to `None` or set to 0 the download is performed directly without streaming.\n* `journal(journal_id: int) -> Journal`<br/>\n  Given a journal ID, it returns a `Journal` object containing the various metadata of the journal.\n* `user(user: str) -> User`<br/>\n  Given a username, it returns a `User` object containing information regarding the user.\n* `gallery(user: str, page: int = 1) -> tuple[list[SubmissionPartial], int | None]`<br/>\n  Returns the list of submissions found on a specific gallery page, and the number of the next page. The returned page\n  number is set to `None` if it is the last page.\n* `scraps(user: str, page: int = 1) -> -> tuple[list[SubmissionPartial], int | None]`<br/>\n  Returns the list of submissions found on a specific scraps page, and the number of the next page. The returned page\n  number is set to `None` if it is the last page.\n* `favorites(user: str, page: str = "") -> tuple[list[SubmissionPartial], str | None]`<br/>\n  Downloads a user\'s favorites page. Because of how favorites pages work on FA, the `page` argument (and the one\n  returned) are strings. If the favorites page is the last then a `None` is returned as next page. An empty page\n  value as argument is equivalent to page 1.<br/>\n  *Note:* favorites page "numbers" do not follow any scheme and are only generated server-side.\n* `journals(user: str, page: int = 1) -> -> tuple[list[JournalPartial], int | None]`<br/>\n  Returns the list of submissions found on a specific journals page, and the number of the next page. The returned page\n  number is set to `None` if it is the last page.\n* `watchlist_to(self, user: str, page:int = 1) -> tuple[list[UserPartial], int | None]`<br/>\n  Given a username, returns a list of `UserPartial` objects for each user that is watching the given user and the next\n  page, if it is not the last, in which case a `None` is returned.\n* `watchlist_by(self, user: str, page:int = 1) -> tuple[list[UserPartial], int | None]`<br/>\n  Given a username, returns a list of `UserPartial` objects for each user that is watched by the given user and the next\n  page, if it is not the last, in which case a `None` is returned.\n\n*Note:* The last page returned by the `watchlist_to` and `watchlist_by` may not be correct as Fur Affinity doesn\'t seem\nto have a consistent behaviour when rendering the next page button, as such it is safer to use an external algorithm to\ncheck whether the method is advancing the page but returning the same/no users.\n\n### UserPartial\n\nA stripped-down class that holds basic user information. It is used to hold metadata gathered when parsing a submission,\njournal, gallery, scraps, etc.\n\n* `name: str` display name with capital letters and extra characters such as "_"\n* `status: str` user status (~, !, etc.)\n* `title: str` the user title as it appears on their userpage\n* `join_date: datetime` the date the user joined (defaults to timestamp 0)\n* `user_tag: bs4.element.Tag` the user element used to parse information (placeholder, `UserPartial` is filled\n  externally)\n\n`UserPartial` objects can be directly cast to a dict object and iterated through.\n\nComparison with `UserPartial` can be made with either another `UserPartial` or `User` object (the URL names are\ncompared), or a string (the URL name is compared to the given string).\n\n#### Init\n\n`__init__(user_tag: bs4.element.Tag = None)`\n\nTo initialise the object, an optional `bs4.element.Tag` object is needed containing the user element from a user page or\nuser folder.\n\nIf no `user_tag` is passed then the object fields will remain at their default - empty - value.\n\n#### Methods\n\n* `name_url -> str`<br/>\n  Property method that returns the URL-safe username\n* `url -> str`<br/>\n  Property method that returns the Fur Affinity URL to the user (`https://www.furaffinity.net/user/{name_url}`).\n* `generate_user_icon_url() -> str`<br/>\n  Generates the URl for the current user icon.\n* `parse(user_page: bs4.BeautifulSoup = None)`<br/>\n  Parses the stored user page for metadata. If `user_page` is passed, it overwrites the existing `user_page` value.\n\n### User\n\nThe main class storing all of a user\'s metadata.\n\n* `name: str` display name with capital letters and extra characters such as "_"\n* `status: str` user status (~, !, etc.)\n* `title: str` the user title as it appears on their userpage\n* `join_date: datetime` the date the user joined (defaults to timestamp 0)\n* `profile: str` profile text in HTML format\n* `profile_bbcode: str` profile text in BBCode format\n* `stats: UserStats` user statistics sorted in a `namedtuple` (`views`, `submissions`, `favorites`, `comments_earned`\n  , `comments_made`, `journals`, `watched_by`, `watching`)\n* `info: dict[str, str]` profile information (e.g. "Accepting Trades", "Accepting Commissions", "Character Species",\n  etc.)\n* `contacts: dict[str, str]` contact links (e.g. Twitter, Steam, etc.)\n* `user_icon_url: str` the URL to the user icon\n* `watched: bool` `True` if the user is watched, `False` otherwise\n* `watched_toggle_link: str | None` The link to toggle the watch status (`/watch/` or `/unwatch/` type link)\n* `blocked: bool` `True` if the user is blocked, `False` otherwise\n* `blocked_toggle_link: str | None` The link to toggle the block status (`/block/` or `/unblock/` type link)\n* `user_page: bs4.BeautifulSoup` the user page used to parse the object fields\n\n`User` objects can be directly cast to a dict object and iterated through.\n\nComparison with `User` can be made with either another `User` or `UserPartial` object (the URL names are compared), or a\nstring (the URL name is compared to the given string).\n\n#### Init\n\n`__init__(user_page: bs4.BeautifulSoup = None)`\n\nTo initialise the object, an optional `bs4.BeautifulSoup` object is needed containing the parsed HTML of a submission\npage.\n\nIf no `user_page` is passed then the object fields will remain at their default - empty - value.\n\n#### Methods\n\n* `name_url -> str`<br/>\n  Property method that returns the URL-safe username\n* `url -> str`<br/>\n  Property method that returns the Fur Affinity URL to the user (`https://www.furaffinity.net/user/{name_url}`).\n* `generate_user_icon_url() -> str`<br/>\n  Generates the URl for the current user icon.\n* `parse(user_page: bs4.BeautifulSoup = None)`<br/>\n  Parses the stored user page for metadata. If `user_page` is passed, it overwrites the existing `user_page` value.\n\n### JournalPartial\n\nThis object contains partial information gathered when parsing a journals folder. It contains the following fields:\n\n* `id: int` journal ID\n* `title: str` journal title\n* `date: datetime` upload date as a [`datetime` object](https://docs.python.org/3/library/datetime.html) (defaults to\n  timestamp 0)\n* `author: UserPartial` journal author (filled only if the journal is parsed from a `bs4.BeautifulSoup` page)\n* `stats: JournalStats` journal statistics stored in a named tuple (`comments` (count))\n* `content: str` journal content in HTML format\n* `content_bbcode: str` journal content in BBCode format\n* `mentions: list[str]` the users mentioned in the content (if they were mentioned as links, e.g. `:iconusername:`,\n  `@username`, etc.)\n* `journal_tag: bs4.element.Tag` the journal tag used to parse the object fields\n\n`JournalPartial` objects can be directly cast to a dict object or iterated through.\n\nComparison with `JournalPartial` can be made with either another `JournalPartial` or `Journal` object (the IDs are\ncompared), or an integer (the `JournalPartial.id` value is compared to the given integer).\n\n#### Init\n\n`__init__(journal_tag: bs4.element.Tag = None)`\n\n`Journal` takes one optional parameters: a journal section tag from a journals page.\n\nIf no `journal_tag` is passed then the object fields will remain at their default - empty - value.\n\n#### Methods\n\n* `url -> str`<br/>\n  Property method that returns the Fur Affinity URL to the journal (`https://www.furaffinity.net/journal/{id}`).\n* `parse(journal_item: bs4.element.Tag = None)`<br/>\n  Parses the stored journal tag for information. If `journal_tag` is passed, it overwrites the existing `journal_tag`\n  value.\n\n### Journal\n\nThis object contains full information gathered when parsing a journal page. It contains the same fields\nas `JournalPartial` with the addition of comments:\n\n* `id: int` journal ID\n* `title: str` journal title\n* `date: datetime` upload date as a [`datetime` object](https://docs.python.org/3/library/datetime.html) (defaults to\n  timestamp 0)\n* `author: UserPartial` journal author (filled only if the journal is parsed from a `bs4.BeautifulSoup` page)\n* `stats: JournalStats` journal statistics stored in a named tuple (`comments` (count))\n* `content: str` journal content in HTML format\n* `content_bbcode: str` journal content in BBCode format\n* `header: str` journal header in HTML format (if present)\n* `footer: str` journal footer in HTML format (if present)\n* `mentions: list[str]` the users mentioned in the content (if they were mentioned as links, e.g. `:iconusername:`,\n  `@username`, etc.)\n* `comments: list[Comments]` the comments to the journal, organised in a tree structure\n* `journal_page: bs4.BeautifulSoup` the journal page used to parse the object fields\n\n`Journal` objects can be directly cast to a dict object or iterated through.\n\nComparison with `Journal` can be made with either another `Journal` or `JournalPartial` object (the IDs are compared),\nor an integer (the `Journal.id` value is compared to the given integer).\n\n#### Init\n\n`__init__(journal_page: bs4.BeautifulSoup = None)`\n\n`Journal` takes one optional journal page argument.\n\nIf no `journal_page` is passed then the object fields will remain at their default - empty - value.\n\n#### Methods\n\n* `url -> str`<br/>\n  Property method that returns the Fur Affinity URL to the journal (`https://www.furaffinity.net/journal/{id}`).\n* `parse(journal_page: bs4.BeautifulSoup = None)`<br/>\n  Parses the stored journal tag for information. If `journal_tag` is passed, it overwrites the existing `journal_tag`\n  value.\n\n### SubmissionPartial\n\nThis lightweight submission object is used to contain the information gathered when parsing gallery, scraps, and\nfavorites pages. It contains only the following fields:\n\n* `id: int` submission ID\n* `title: str` submission title\n* `author: UserPartial` submission author (only the `name` field is filled)\n* `rating: str` submission rating [general, mature, adult]\n* `type: str` submission type [text, image, etc...]\n* `thumbnail_url: str` the URL to the submission thumbnail\n* `submission_figure: bs4.element.Tag` the figure tag used to parse the object fields\n\n`SubmissionPartial` objects can be directly cast to a dict object or iterated through.\n\nComparison with `Submission` can be made with either another `SubmissionPartial` or `Submission` object (the IDs are\ncompared), or an integer (the `Submission.id` value is compared to the given integer).\n\n#### Init\n\n`__init__(submission_figure: bs4.element.Tag = None)`\n\nTo initialise the object, an optional `bs4.element.Tag` object is needed containing the parsed HTML of a submission\nfigure tag.\n\nIf no `submission_figure` is passed then the object fields will remain at their default - empty - value.\n\n#### Methods\n\n* `url -> str`<br/>\n  Property method that returns the Fur Affinity URL to the submission (`https://www.furaffinity.net/view/{id}`).\n* `parse(submission_figure: bs4.element.Tag = None)`<br/>\n  Parses the stored submission figure tag for information. If `submission_figure` is passed, it overwrites the\n  existing `submission_figure` value.\n\n### Submission\n\nThe main class that parses and holds submission metadata.\n\n* `id: int` submission ID\n* `title: str` submission title\n* `author: UserPartial` submission author (only the `name`, `title`, and `user_icon_url` fields are filled)\n* `date: datetime` upload date as a [`datetime` object](https://docs.python.org/3/library/datetime.html) (defaults to\n  timestamp 0)\n* `tags: list[str]` tags list\n* `category: str` category\n* `species: str` species\n* `gender: str` gender\n* `rating: str` rating\n* `stats: SubmissionStats` submission statistics stored in a named tuple (`views`, `comments` (count), `favorites`)\n* `type: str` submission type (text, image, etc...)\n* `description: str` description in HTML format\n* `description_bbcode: str` description in BBCode format\n* `footer: str` footer in HTML format\n* `mentions: list[str]` the users mentioned in the description (if they were mentioned as links, e.g. `:iconusername:`,\n  `@username`, etc.)\n* `folder: str` the submission folder (gallery or scraps)\n* `user_folders: list[SubmissionUserFolder]` user folders stored in a list of named tuples (`name`, `url`, `group` (\n  if any))\n* `file_url: str` the URL to the submission file\n* `thumbnail_url: str` the URL to the submission thumbnail\n* `prev: int` the ID of the previous submission (if any)\n* `next: int` the ID of the next submission (if any)\n* `favorite: bool` `True` if the submission is a favorite, `False` otherwise\n* `favorite_toggle_link: str` the link to toggle the favorite status (`/fav/` or `/unfav/` type URL)\n* `comments: list[Comments]` the comments to the submission, organised in a tree structure\n* `submission_page: bs4.BeautifulSoup` the submission page used to parse the object fields\n\n`Submission` objects can be directly cast to a dict object and iterated through.\n\nComparison with `Submission` can be made with either another `Submission` or `SubmissionPartial` object (the IDs are\ncompared), or an integer (the `Submission.id` value is compared to the given integer).\n\n#### Init\n\n`__init__(submission_page: bs4.BeautifulSoup = None)`\n\nTo initialise the object, an optional `bs4.BeautifulSoup` object is needed containing the parsed HTML of a submission\npage.\n\nIf no `submission_page` is passed then the object fields will remain at their default - empty - value.\n\n#### Methods\n\n* `url -> str`<br/>\n  Property method that returns the Fur Affinity URL to the submission (`https://www.furaffinity.net/view/{id}`).\n* `parse(submission_page: bs4.BeautifulSoup = None)`<br/>\n  Parses the stored submission page for metadata. If `submission_page` is passed, it overwrites the\n  existing `submission_page` value.\n\n### Comment\n\nThis object class contains comment metadata and is used to build a tree structure with the comments and their replies.\n\n* `id: int` the comment ID\n* `author: UserPartial` the user who posted the comment\n* `date: datetime` the date the comment was posted\n* `text: str` the comment text in HTML format\n* `text_bbcode: str` the comment text in BBCode format\n* `replies: list[Comment]` list of replies to the comment\n* `reply_to: Comment | int | None` the parent comment, if the comment is a reply. The variable type is `int` only if the\n  comment is parsed outside the parse method of a `Submission` or `Journal` (e.g. by creating a new comment with a\n  comment tag), and when iterating over the parent object (to avoid infinite recursion errors), be it `Submission`\n  , `Journal` or another `Comment`.\n* `edited: bool` `True` if the comment was edited, `False` otherwise\n* `hidden: bool` `True` if the comment was hidden, `False` otherwise (if the comment was hidden, the author and date\n  fields will default to their empty values)\n* `parent: Submission | Journal | None` the `Submission` or `Journal` object the comments are connected to\n* `comment_tag: bs4.element.Tag` the comment tag used to parse the object fields\n\n`Comment` objects can be directly cast to a dict object and iterated through.\n\nComparison with `Comment` can be made with either another comment (the IDs are compared), or an integer (\nthe `Comment.id` value is compared to the given integer).\n\n*Note:* The `__iter__` method of `Comment` objects automatically removes recursion. The `parent` variable is set\nto `None` and `reply_to` is set to the comment\'s ID.<br/>\n*Note:* Because each comment contains the parent `Submission` or `Journal` object (which contains the comment itself)\nand the replied comment object, some iterations may cause infinite recursion errors, for example when using\nthe `copy.deepcopy` function. If such iterations are needed, simply set the `parent` variable to `None` and\nthe `reply_to` variable to `None` or the comment\'s ID (this can be done easily after flattening the comments list\nwith `faapi.comment.flatten_comments`, the comments can then be sorted again with `faapi.comment.sort_comments` which\nwill also restore the `reply_to` values to `Comment` objects).\n\n#### Init\n\n`__init__(self, tag: bs4.element.Tag = None, parent: Submission | Journal = None)`\n\nTo initialise the object, an optional `bs4.element.Tag` object is needed containing the comment tag as taken from a\nsubmission/journal page.\n\nThe optional `parent` argument sets the `parent` variable described above.\n\nIf no `tag` is passed then the object fields will remain at their default - empty - value.\n\n#### Methods\n\n* `url -> str`<br/>\n  Property method that returns the Fur Affinity URL to the comment (\n  e.g. `https://www.furaffinity.net/view/12345678#cid:1234567890`). If the `parent` variable is `None`, the property\n  returns an empty string.\n* `parse(tag: bs4.element.Tag = None)`<br/>\n  Parses the stored tag for metadata. If `tag` is passed, it overwrites the existing `tag` value.\n\n#### Extra Functions\n\nThese extra functions can be used to operate on a list of comments. They only alter the order and structure, but they do\nnot touch any of the metadata.\n\n* `faapi.comment.sort_comments(comments: list[Comment]) -> list[Comment]`<br/>\n  Sorts a list of comments into a tree structure. Replies are overwritten.\n* `faapi.comment.flatten_comments(comments: list[Comment]) -> list[Comment]`<br/>\n  Flattens a list of comments. Replies are not modified.\n\n#### Comment Tree Graphs\n\nUsing the tree structure generated by the library, it is trivial to build a graph visualisation of the comment tree\nusing the [DOT](https://www.graphviz.org/doc/info/lang.html) language.\n\n```python\nsubmission, _ = api.submission(12345678)\ncomments = faapi.comment.flatten_comments(submission.comments)\nwith open("comments.dot", "w") as f:\n    f.write("digraph {\\n")\n    for comment in [c for c in comments if c.reply_to is None]:\n        f.write(f"    parent -> {comment.id}\\n")\n    for comment in comments:\n        for reply in comment.replies:\n            f.write(f"    {comment.id} -> {reply.id}\\n")\n    f.write("}")\n```\n\n```dot\ndigraph {\n    parent -> 157990848\n    parent -> 157993838\n    parent -> 157997294\n    157990848 -> 158014077\n    158014077 -> 158014816\n    158014816 -> 158093180\n    158093180 -> 158097024\n    157993838 -> 157998464\n    157993838 -> 158014126\n    157997294 -> 158014135\n    158014135 -> 158014470\n    158014135 -> 158030074\n    158014470 -> 158093185\n    158030074 -> 158093199\n}\n```\n\n<img alt="comments tree graph" width="400" src="https://quickchart.io/graphviz?graph=digraph%7B%0Aparent-%3E157990848%0Aparent-%3E157993838%0Aparent-%3E157997294%0A157990848-%3E158014077%0A158014077-%3E158014816%0A158014816-%3E158093180%0A158093180-%3E158097024%0A157993838-%3E157998464%0A157993838-%3E158014126%0A157997294-%3E158014135%0A158014135-%3E158014470%0A158014135-%3E158030074%0A158014470-%3E158093185%0A158030074-%3E158093199%0A%7D">\n\n_The graph above was generated with [quickchart.io](https://quickchart.io/documentation/graphviz-api/)_\n\n## BBCode Conversion\n\nUsing the BBCode fields allows to convert between the raw HTMl recovered from Fur Affinity and BBCode tags that follow\nFA\'s guidelines. Conversion from HTML to BBCode covers all known tags and preserves all newlines and spacing.\n\nBBCode text can be converted to Fur Affinity\'s HTMl using the `faapi.parse.bbcode_to_html()` function. The majority of\nsubmissions can be converted back and forth between HTML and BBCode without any information loss, however, the parser\nrules are still a work in progress and there are many edge cases where unusual text and formatting cause the parser to\ngenerate incorrect HTML.\n\n## Exceptions\n\nThe following are the exceptions explicitly raised by the FAAPI functions. The exceptions deriving from `ParsingError`\nare chosen depending on the content of the page. Because Fur Affinity doesn\'t use HTTP status codes besides 404, the\npage is checked against a static list of known error messages/page titles in order to determine the specific error to be\nused. If no match is found, then the `ServerError` (if the page has the "Server Error" title) or the more\ngeneral `NoticeMessage` exceptions are used instead. The actual error message parsed from the page is used as argument\nfor the exceptions, so that it can be analysed when caught.\n\n* `DisallowedPath(Exception)` The path is not allowed by the robots.txt.\n* `Unauthorized(Exception)` The user is not logged-in.\n* `ParsingError(Exception)` An error occurred while parsing the page.\n    * `NonePage(ParsingError)` The parsed page is `None`.\n    * `NotFound(ParsingError)` The resource could not be found (general 404 page or non-existing submission, user, or\n      journal).\n    * `NoTitle(ParsingError)` The parsed paged is missing a title.\n    * `DisabledAccount(ParsingError)` The resource belongs to a disabled account.\n    * `ServerError(ParsingError)` The page contains a server error notice.\n    * `NoticeMessage(ParsingError)` A notice of unknown type was found in the page.\n\n## Beautiful Soup Warnings\n\nWhen parsing some pages or converting HTML to BBCode, the [Beautiful Soup](https://pypi.org/project/beautifulsoup4/)\nlibrary may give some warnings, for example `MarkupResemblesLocatorWarning`. These warnings are left enabled for\nclarity, but can be disabled manually using the `warnings.filterwarnings` function.\n\n## Contributing\n\nAll contributions and suggestions are welcome!\n\nIf you have suggestions for fixes or improvements, you can open an issue with your idea, see [#Issues](#issues) for\ndetails.\n\n## Issues\n\nIf any problem is encountered during usage of the program, an issue can be opened\non [GitHub](https://github.com/FurryCoders/FAAPI/issues).\n\nIssues can also be used to suggest improvements and features.\n\nWhen opening an issue for a problem, please copy the error message and describe the operation in progress when the error\noccurred.\n',
    'author': 'Matteo Campinoti',
    'author_email': 'matteo.campinoti94@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/FurryCoders/FAAPI',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from collections import namedtuple
from datetime import datetime
from typing import Optional

from .connection import join_url
from .connection import root
from .exceptions import _raise_exception
from .parse import BeautifulSoup
from .parse import Tag
from .parse import check_page_raise
from .parse import html_to_bbcode
from .parse import parse_user_page
from .parse import parse_user_tag
from .parse import username_url


class UserStats(namedtuple("UserStats", ["views", "submissions", "favorites", "comments_earned",
                                         "comments_made", "journals", "watched_by", "watching"])):
    """
    This object contains a user's statistics:
    * views
    * submissions
    * favorites
    * comments_earned
    * comments_made
    * journals
    * watched_by
    * watching
    """


class UserBase:
    """
    Base class for the user objects.
    """

    def __init__(self):
        self.name: str = ""
        self.status: str = ""

    def __hash__(self) -> int:
        return hash(self.name_url)

    def __eq__(self, other) -> bool:
        if isinstance(other, UserBase):
            return other.name_url == self.name_url
        elif isinstance(other, str):
            return username_url(other) == self.name_url
        return False

    def __gt__(self, other) -> bool:
        if isinstance(other, UserBase):
            return self.name_url > other.name_url
        elif isinstance(other, str):
            return self.name_url > username_url(other)
        return False

    def __ge__(self, other) -> bool:
        if isinstance(other, UserBase):
            return self.name_url >= other.name_url
        elif isinstance(other, str):
            return self.name_url >= username_url(other)
        return False

    def __lt__(self, other) -> bool:
        if isinstance(other, UserBase):
            return self.name_url < other.name_url
        elif isinstance(other, str):
            return self.name_url < username_url(other)
        return False

    def __le__(self, other) -> bool:
        if isinstance(other, UserBase):
            return self.name_url <= other.name_url
        elif isinstance(other, str):
            return self.name_url <= username_url(other)
        return False

    def __iter__(self):
        yield "name", self.name
        yield "status", self.status

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return self.status + self.name

    @property
    def name_url(self):
        """
        Compose the URL-safe username.

        :return: The cleaned username.
        """
        return username_url(self.name)

    @property
    def url(self):
        """
        Compose the full URL to the user.

        :return: The URL to the user.
        """
        return join_url(root, "user", self.name_url)

    def generate_user_icon_url(self) -> str:
        """
        Generate the URl for the current user icon.

        :return: The URL to the user icon
        """
        return f"https://a.furaffinity.net/{datetime.now():%Y%m%d}/{self.name_url}.gif"


class UserPartial(UserBase):
    """
    Contains partial user information gathered from user folders (gallery, journals, etc.) and submission/journal pages.
    """

    def __init__(self, user_tag: Tag = None):
        """
        :param user_tag: The tag from which to parse the user information.
        """
        assert user_tag is None or isinstance(user_tag, Tag), \
            _raise_exception(TypeError(f"user_tag must be {None} or {Tag.__name__}"))

        super().__init__()

        self.user_tag: Optional[Tag] = user_tag
        self.title: str = ""
        self.join_date: datetime = datetime.fromtimestamp(0)
        self.user_icon_url: str = ""

        self.parse()

    def __iter__(self):
        yield "name", self.name
        yield "status", self.status
        yield "title", self.title
        yield "join_date", self.join_date
        yield "user_icon_url", self.user_icon_url

    def parse(self, user_tag: Tag = None):
        """
        Parse a user page, overrides any information already present in the object.

        :param user_tag: The tag from which to parse the user information.
        """
        assert user_tag is None or isinstance(user_tag, Tag), \
            _raise_exception(TypeError(f"user_tag must be {None} or {Tag.__name__}"))

        self.user_tag = user_tag or self.user_tag
        if self.user_tag is None:
            return

        parsed: dict = parse_user_tag(self.user_tag)

        self.name = parsed["name"]
        self.status = parsed["status"]
        self.title = parsed["title"]
        self.join_date = parsed["join_date"]


class User(UserBase):
    """
    Contains complete user information gathered from userpages.
    """

    def __init__(self, user_page: BeautifulSoup = None):
        """
        :param user_page: The page from which to parse the user information.
        """
        assert user_page is None or isinstance(user_page, BeautifulSoup), \
            _raise_exception(TypeError(f"user_page must be {None} or {BeautifulSoup.__name__}"))

        super().__init__()

        self.user_page: Optional[BeautifulSoup] = user_page
        self.title: str = ""
        self.join_date: datetime = datetime.fromtimestamp(0)
        self.profile: str = ""
        self.stats: UserStats = UserStats(0, 0, 0, 0, 0, 0, 0, 0)
        self.info: dict[str, str] = {}
        self.contacts: dict[str, str] = {}
        self.user_icon_url: str = ""
        self.watched: bool = False
        self.watched_toggle_link: Optional[str] = None
        self.blocked: bool = False
        self.blocked_toggle_link: Optional[str] = None

        self.parse()

    def __iter__(self):
        yield "name", self.name
        yield "status", self.status
        yield "title", self.title
        yield "join_date", self.join_date
        yield "profile", self.profile
        yield "stats", self.stats._asdict()
        yield "info", self.info
        yield "contacts", self.contacts
        yield "user_icon_url", self.user_icon_url
        yield "watched", self.watched
        yield "watched_toggle_link", self.watched_toggle_link
        yield "blocked", self.blocked
        yield "blocked_toggle_link", self.blocked_toggle_link

    @property
    def profile_bbcode(self) -> str:
        """
        The user profile text formatted to BBCode

        :return: BBCode profile
        """
        return html_to_bbcode(self.profile)

    def parse(self, user_page: BeautifulSoup = None):
        """
        Parse a user page, overrides any information already present in the object.

        :param user_page: The page from which to parse the user information.
        """
        assert user_page is None or isinstance(user_page, BeautifulSoup), \
            _raise_exception(TypeError(f"user_page must be {None} or {BeautifulSoup.__name__}"))

        self.user_page = user_page or self.user_page
        if self.user_page is None:
            return

        check_page_raise(self.user_page)

        parsed: dict = parse_user_page(self.user_page)

        self.name = parsed["name"]
        self.status = parsed["status"]
        self.profile = parsed["profile"]
        self.title = parsed["title"]
        self.join_date = parsed["join_date"]
        self.stats = UserStats(*parsed["stats"])
        self.info = parsed["info"]
        self.contacts = parsed["contacts"]
        self.user_icon_url = parsed["user_icon_url"]
        self.watched = parsed["watch"] is None and parsed["unwatch"] is not None
        self.watched_toggle_link = parsed["watch"] or parsed["unwatch"] or None
        self.blocked = parsed["block"] is None and parsed["unblock"] is not None
        self.blocked_toggle_link = parsed["block"] or parsed["unblock"] or None
